"""
RunPod | CLI | Project | Functions
"""

import os
import sys
import uuid
from datetime import datetime

import tomlkit
from tomlkit import comment, document, nl, table

from runpod import (
    __version__,
    create_endpoint,
    create_template,
    get_pod,
    update_endpoint_template,
)
from runpod.cli import BASE_DOCKER_IMAGE, ENV_VARS, GPU_TYPES
from runpod.cli.utils.ssh_cmd import SSHConnection

from ...utils.rp_sync import sync_directory
from .helpers import (
    attempt_pod_launch,
    copy_template_files,
    get_project_endpoint,
    get_project_pod,
    load_project_config,
)

STARTER_TEMPLATES = os.path.join(os.path.dirname(__file__), "starter_templates")


def _launch_dev_pod():
    """Launch a development pod."""
    config = load_project_config()  # Load runpod.toml

    print("Deploying development pod on RunPod...")

    # Prepare the environment variables
    environment_variables = {"RUNPOD_PROJECT_ID": config["project"]["uuid"]}
    for variable in config["project"].get("env_vars", {}):
        environment_variables[variable] = config["project"]["env_vars"][variable]

    # Prepare the GPU types
    selected_gpu_types = config["project"].get("gpu_types", [])
    if config["project"].get("gpu", None):
        selected_gpu_types.append(config["project"]["gpu"])

    # Attempt to launch a pod with the given configuration
    new_pod = attempt_pod_launch(config, environment_variables)
    if new_pod is None:
        print(
            "Selected GPU types unavailable, try again later or use a different type."
        )
        return None

    print("Waiting for pod to come online... ", end="")
    sys.stdout.flush()

    # Wait for the pod to come online
    while (
        new_pod.get("desiredStatus", None) != "RUNNING"
        or new_pod.get("runtime") is None
    ):
        new_pod = get_pod(new_pod["id"])

    project_pod_id = new_pod["id"]

    print(
        f"Project {config['project']['name']} pod ({project_pod_id}) created.",
        end="\n\n",
    )
    return project_pod_id


# -------------------------------- New Project ------------------------------- #
def create_new_project(
    project_name,
    runpod_volume_id,
    cuda_version,
    python_version,  # pylint: disable=too-many-locals, too-many-arguments, too-many-statements
    model_type=None,
    model_name=None,
    init_current_dir=False,
):
    """Create a new project."""
    if not init_current_dir:
        project_folder = os.path.join(os.getcwd(), project_name)
        if not os.path.exists(project_folder):
            os.makedirs(project_folder)

        if model_type is None:
            model_type = "default"

        template_dir = os.path.join(STARTER_TEMPLATES, model_type)

        copy_template_files(template_dir, project_folder)

        # Replace placeholders in requirements.txt
        requirements_path = os.path.join(project_folder, "builder/requirements.txt")
        with open(requirements_path, "r", encoding="utf-8") as requirements_file:
            requirements_content = requirements_file.read()

        if "dev" in __version__:
            requirements_content = requirements_content.replace(
                "<<RUNPOD>>", "git+https://github.com/runpod/runpod-python.git"
            )
        else:
            requirements_content = requirements_content.replace(
                "<<RUNPOD>>", f"runpod=={__version__}"
            )

        with open(requirements_path, "w", encoding="utf-8") as requirements_file:
            requirements_file.write(requirements_content)

        # If there's a model_name, replace placeholders in handler.py
        if model_name:
            handler_path = os.path.join(project_name, "src/handler.py")
            with open(handler_path, "r", encoding="utf-8") as file:
                handler_content = file.read()
                handler_content = handler_content.replace("<<MODEL_NAME>>", model_name)

            with open(handler_path, "w", encoding="utf-8") as file:
                file.write(handler_content)
    else:
        project_folder = os.getcwd()

    project_uuid = str(uuid.uuid4())[:8]

    toml_config = document()
    toml_config.add(comment("RunPod Project Configuration"))
    toml_config.add(nl())
    toml_config.add("title", project_name)

    project_table = table()
    project_table.add("uuid", project_uuid)
    project_table.add("name", project_name)
    project_table.add("base_image", BASE_DOCKER_IMAGE.format(cuda_version=cuda_version))
    project_table.add("gpu_types", GPU_TYPES)
    project_table.add("gpu_count", 1)
    project_table.add("storage_id", runpod_volume_id)
    project_table.add("volume_mount_path", "/runpod-volume")
    project_table.add("ports", "8080/http, 22/tcp")
    project_table.add("container_disk_size_gb", 10)
    project_table.add("env_vars", ENV_VARS)
    toml_config.add("project", project_table)

    template_table = table()
    template_table.add("model_type", str(model_type))
    template_table.add("model_name", str(model_name))
    toml_config.add("template", template_table)

    runtime_table = table()
    runtime_table.add("python_version", python_version)
    runtime_table.add("handler_path", "src/handler.py")
    runtime_table.add("requirements_path", "builder/requirements.txt")
    toml_config.add("runtime", runtime_table)

    with open(
        os.path.join(project_folder, "runpod.toml"), "w", encoding="UTF-8"
    ) as config_file:
        tomlkit.dump(toml_config, config_file)


# ------------------------------- Start Project ------------------------------ #
def start_project():  # pylint: disable=too-many-locals, too-many-branches
    """
    Start the project development environment from runpod.toml
    - Check if the project pod already exists.

    - If the project pod does not exist:
        # SSH into the pod and create a project folder within the volume
        # Create a folder in the volume for the project that matches the project uuid
        # Create a folder in the project folder that matches the project name
        # Copy the project files into the project folder
        # crate a virtual environment using the python version specified in the project config
        # install the requirements.txt file

    - If the project pod does exist:

    """
    config = load_project_config()  # Load runpod.toml

    for config_item in config["project"]:
        print(f'    - {config_item}: {config["project"][config_item]}')
    print("")

    project_pod_id = get_project_pod(config["project"]["uuid"])

    # Check if the project pod already exists, if not create it.
    if not project_pod_id:
        project_pod_id = _launch_dev_pod()

    if project_pod_id is None:
        return

    with SSHConnection(project_pod_id) as ssh_conn:

        project_path_uuid = (
            f'{config["project"]["volume_mount_path"]}/{config["project"]["uuid"]}'
        )
        project_path_uuid_dev = os.path.join(project_path_uuid, "dev")
        project_path_uuid_prod = os.path.join(project_path_uuid, "prod")
        remote_project_path = os.path.join(
            project_path_uuid_dev, config["project"]["name"]
        )

        # Create the project folder on the pod
        print(
            f"Checking pod project folder: {remote_project_path} on pod {project_pod_id}"
        )
        ssh_conn.run_commands(
            [f"mkdir -p {remote_project_path} {project_path_uuid_prod}"]
        )

        # Copy local files to the pod project folder
        print(f"Syncing files to pod {project_pod_id}")
        ssh_conn.rsync(os.getcwd(), project_path_uuid_dev)

        # Create the virtual environment
        venv_path = os.path.join(project_path_uuid_dev, "venv")
        print(
            f"Activating Python virtual environment: {venv_path} on pod {project_pod_id}"
        )
        commands = [
            f'python{config["runtime"]["python_version"]} -m virtualenv {venv_path}',
            f"source {venv_path}/bin/activate &&"
            f"cd {remote_project_path} &&"
            "python -m pip install --upgrade pip &&"
            f'python -m pip install -v --requirement {config["runtime"]["requirements_path"]}',
        ]
        ssh_conn.run_commands(commands)

        # Start the watcher and then start the API development server
        sync_directory(ssh_conn, os.getcwd(), project_path_uuid_dev)

        project_name = config["project"]["name"]
        pip_req_path = os.path.join(
            remote_project_path, config["runtime"]["requirements_path"]
        )
        handler_path = os.path.join(
            remote_project_path, config["runtime"]["handler_path"]
        )

        launch_api_server = [
            f"""
            pkill inotify

            function force_kill {{
                kill $1 2>/dev/null
                sleep 1

                if ps -p $1 > /dev/null; then
                    echo "Graceful kill failed, attempting SIGKILL..."
                    kill -9 $1 2>/dev/null
                    sleep 1

                    if ps -p $1 > /dev/null; then
                        echo "Failed to kill process with PID: $1"
                        exit 1
                    else
                        echo "Killed process with PID: $1 using SIGKILL"
                    fi

                else
                    echo "Killed process with PID: $1"
                fi
            }}

            function cleanup {{
                echo "Cleaning up..."
                force_kill $last_pid
            }}

            trap cleanup EXIT SIGINT

            if source {project_path_uuid_dev}/venv/bin/activate; then
                echo -e "- Activated virtual environment."
            else
                echo "Failed to activate virtual environment."
                exit 1
            fi

            if cd {project_path_uuid_dev}/{project_name}; then
                echo -e "- Changed to project directory."
            else
                echo "Failed to change directory."
                exit 1
            fi

            exclude_pattern='(__pycache__|\\.pyc$)'
            function update_exclude_pattern {{
                exclude_pattern='(__pycache__|\\.pyc$)'
                if [[ -f .runpodignore ]]; then
                    while IFS= read -r line; do
                        line=$(echo "$line" | tr -d '[:space:]')
                        [[ "$line" =~ ^#.*$ || -z "$line" ]] && continue # Skip comments and empty lines
                        exclude_pattern="${{exclude_pattern}}|(${{line}})"
                    done < .runpodignore
                    echo -e "- Ignoring files matching pattern: $exclude_pattern"
                fi
            }}
            update_exclude_pattern

            # Start the API server in the background, and save the PID
            python {handler_path} --rp_serve_api --rp_api_host="0.0.0.0" --rp_api_port=8080 --rp_api_concurrency=1 &
            last_pid=$!

            echo -e "- Started API server with PID: $last_pid" && echo ""
            echo "Connect to the API server at:"
            echo ">  https://$RUNPOD_POD_ID-8080.proxy.runpod.net" && echo ""

            while true; do
                if changed_file=$(inotifywait -q -r -e modify,create,delete --exclude "$exclude_pattern" {remote_project_path} --format '%w%f'); then
                    echo "Detected changes in: $changed_file"
                else
                    echo "Failed to detect changes."
                    exit 1
                fi

                force_kill $last_pid

                if [[ $changed_file == *"requirements"* ]]; then
                    echo "Installing new requirements..."
                    python -m pip install --upgrade pip && python -m pip install -r {pip_req_path}
                fi

                if [[ $changed_file == *".runpodignore"* ]]; then
                    update_exclude_pattern
                fi

                python {handler_path} --rp_serve_api --rp_api_host="0.0.0.0" --rp_api_port=8080 --rp_api_concurrency=1 &
                last_pid=$!

                echo "Restarted API server with PID: $last_pid"
            done
        """
        ]

        print("")
        print("Starting project development endpoint...")
        ssh_conn.run_commands(launch_api_server)


# ------------------------------ Deploy Project ------------------------------ #
def create_project_endpoint():
    """Create a project endpoint.
    - Move code in dev to prod folder
    - TODO: git commit the diff from current state to new state
    - Create a serverless template for the project
    - Create a new endpoint using the template
    """
    config = load_project_config()
    project_pod_id = get_project_pod(config["project"]["uuid"])

    # Check if the project pod already exists, if not create it.
    if not project_pod_id:
        project_pod_id = _launch_dev_pod()

    if project_pod_id is None:
        return None

    with SSHConnection(project_pod_id) as ssh_conn:
        project_path_uuid = (
            f'{config["project"]["volume_mount_path"]}/{config["project"]["uuid"]}'
        )
        project_path_uuid_prod = os.path.join(project_path_uuid, "prod")
        remote_project_path = os.path.join(
            project_path_uuid_prod, config["project"]["name"]
        )

        # Copy local files to the pod project folder
        ssh_conn.run_commands([f"mkdir -p {remote_project_path}"])
        print(f"Syncing files to pod {project_pod_id} prod")
        ssh_conn.rsync(os.getcwd(), project_path_uuid_prod)

        # Create the virtual environment
        venv_path = os.path.join(project_path_uuid_prod, "venv")
        print(
            f"Activating Python virtual environment: {venv_path} on pod {project_pod_id}"
        )
        commands = [
            f'python{config["runtime"]["python_version"]} -m venv {venv_path}',
            f"source {venv_path}/bin/activate &&"
            f"cd {remote_project_path} &&"
            "python -m pip install --upgrade pip &&"
            f'python -m pip install -v --requirement {config["runtime"]["requirements_path"]}',
        ]
        ssh_conn.run_commands(commands)
        ssh_conn.close()

    environment_variables = {}
    for variable in config["project"]["env_vars"]:
        environment_variables[variable] = config["project"]["env_vars"][variable]

    # Construct the docker start command
    activate_cmd = (
        f'. /runpod-volume/{config["project"]["uuid"]}/prod/venv/bin/activate'
    )
    python_cmd = f'python -u /runpod-volume/{config["project"]["uuid"]}/prod/{config["project"]["name"]}/{config["runtime"]["handler_path"]}'  # pylint: disable=line-too-long
    docker_start_cmd = 'bash -c "' + activate_cmd + " && " + python_cmd + '"'

    project_endpoint_template = create_template(
        name=f'{config["project"]["name"]}-endpoint | {config["project"]["uuid"]} | {datetime.now()}',  # pylint: disable=line-too-long
        image_name=config["project"]["base_image"],
        container_disk_in_gb=config["project"]["container_disk_size_gb"],
        docker_start_cmd=docker_start_cmd,
        env=environment_variables,
        is_serverless=True,
    )

    deployed_endpoint = get_project_endpoint(config["project"]["uuid"])
    if not deployed_endpoint:
        deployed_endpoint = create_endpoint(
            name=f'{config["project"]["name"]}-endpoint | {config["project"]["uuid"]}',
            template_id=project_endpoint_template["id"],
            network_volume_id=config["project"]["storage_id"],
        )
    else:
        deployed_endpoint = update_endpoint_template(
            endpoint_id=deployed_endpoint["id"],
            template_id=project_endpoint_template["id"],
        )

    # does user want to tear down and recreate workers immediately?

    return deployed_endpoint["id"]



"""
RunPod | CLI | Groups | Project | Commands | Tests
"""

import unittest
from unittest.mock import patch

from click.testing import CliRunner

from runpod.cli.groups.project.commands import (
    deploy_project,
    new_project_wizard,
    start_project_pod,
)


class TestProjectCLI(unittest.TestCase):
    """A collection of tests for the Project CLI commands."""

    def setUp(self):
        self.runner = CliRunner()

    def test_new_project_wizard_no_network_volumes(self):
        """
        Tests the new_project_wizard command with no network volumes.
        """
        with patch("runpod.cli.groups.project.commands.get_user") as mock_get_user:
            mock_get_user.return_value = {"networkVolumes": []}

            result = self.runner.invoke(new_project_wizard)

        self.assertEqual(result.exit_code, 1)
        self.assertIn("You do not have any network volumes.", result.output)

    def test_new_project_wizard_success(self):
        """
        Tests the new_project_wizard command.
        """
        with patch("click.prompt") as mock_prompt, patch(
            "click.confirm", return_value=True
        ) as mock_confirm, patch(
            "runpod.cli.groups.project.commands.create_new_project"
        ) as mock_create, patch(
            "runpod.cli.groups.project.commands.get_user"
        ) as mock_get_user, patch(
            "runpod.cli.groups.project.commands.cli_select"
        ) as mock_select:
            mock_get_user.return_value = {
                "networkVolumes": [
                    {
                        "id": "XYZ_VOLUME",
                        "name": "XYZ_VOLUME",
                        "size": 100,
                        "dataCenterId": "XYZ",
                    }
                ]
            }  # pylint: disable=line-too-long
            mock_prompt.side_effect = ["TestProject", "11.8.0", "3.10"]
            mock_select.return_value = {"volume-id": "XYZ_VOLUME"}

            result = self.runner.invoke(
                new_project_wizard,
                ["--type", "llama2", "--model", "meta-llama/Llama-2-7b"],
            )  # pylint: disable=line-too-long

        self.assertEqual(result.exit_code, 0)
        mock_confirm.assert_called_with("Do you want to continue?", abort=True)
        mock_create.assert_called()
        mock_prompt.assert_called()
        mock_create.assert_called_with(
            "TestProject",
            "XYZ_VOLUME",
            "11.8.0",
            "3.10",
            "llama2",
            "meta-llama/Llama-2-7b",
            False,
        )  # pylint: disable=line-too-long
        self.assertIn("Project TestProject created successfully!", result.output)

    def test_new_project_wizard_success_init_current_dir(self):
        """
        Tests the new_project_wizard command with the --init flag.
        """
        with patch("click.prompt") as mock_prompt, patch(
            "click.confirm", return_value=True
        ) as mock_confirm, patch(
            "runpod.cli.groups.project.commands.create_new_project"
        ) as mock_create, patch(
            "runpod.cli.groups.project.commands.get_user"
        ) as mock_get_user, patch(
            "runpod.cli.groups.project.commands.cli_select"
        ) as mock_select, patch(
            "os.getcwd"
        ) as mock_getcwd:
            mock_get_user.return_value = {
                "networkVolumes": [
                    {
                        "id": "XYZ_VOLUME",
                        "name": "XYZ_VOLUME",
                        "size": 100,
                        "dataCenterId": "XYZ",
                    }
                ]
            }  # pylint: disable=line-too-long
            mock_select.return_value = {"volume-id": "XYZ_VOLUME"}
            mock_prompt.side_effect = ["XYZ_VOLUME", "11.8.0", "3.10"]

            self.runner.invoke(new_project_wizard, ["--init"])
            assert mock_confirm.called
            assert mock_create.called
            assert mock_getcwd.called

    def test_new_project_wizard_invalid_name(self):
        """
        Tests the new_project_wizard command with an invalid project name.
        """
        with patch("runpod.cli.groups.project.commands.get_user") as mock_get_user:
            mock_get_user.return_value = {"networkVolumes": ["XYZ_VOLUME"]}

            result = self.runner.invoke(new_project_wizard, ["--name", "Invalid/Name"])

        self.assertEqual(result.exit_code, 2)
        self.assertIn("Project name contains an invalid character", result.output)

    def test_start_project_pod(self):
        """
        Tests the start_project_pod command.
        """
        with patch("click.confirm", return_value=True) as mock_confirm, patch(
            "runpod.cli.groups.project.commands.start_project"
        ) as mock_start:
            mock_start.return_value = None
            result = self.runner.invoke(start_project_pod)

        mock_confirm.assert_called_with("Do you want to continue?", abort=True)
        self.assertEqual(result.exit_code, 0)
        self.assertIn("Starting project development pod...", result.output)

    @patch("runpod.cli.groups.project.commands.click.echo")
    @patch("runpod.cli.groups.project.commands.create_project_endpoint")
    def test_deploy_project(self, mock_create_project_endpoint, mock_click_echo):
        """Test the deploy_project function."""
        mock_create_project_endpoint.return_value = "test_endpoint_id"

        result = self.runner.invoke(deploy_project)

        mock_create_project_endpoint.assert_called_once()

        mock_click_echo.assert_any_call("Deploying project...")
        mock_click_echo.assert_any_call("The following urls are available:")
        mock_click_echo.assert_any_call(
            "    - https://api.runpod.ai/v2/test_endpoint_id/runsync"
        )
        mock_click_echo.assert_any_call(
            "    - https://api.runpod.ai/v2/test_endpoint_id/run"
        )
        mock_click_echo.assert_any_call(
            "    - https://api.runpod.ai/v2/test_endpoint_id/health"
        )

        self.assertEqual(result.exit_code, 0)



""" runpod-python version """

from importlib.metadata import PackageNotFoundError, version


def get_version():
    """ Get the version of runpod-python """ ""
    try:
        return version("runpod")
    except PackageNotFoundError:
        return "unknown"


__version__ = get_version()



""" Tests for runpod | serverless | modules | download.py """

# pylint: disable=R0903,W0613

import os
import unittest
from unittest.mock import MagicMock, mock_open, patch

from requests import RequestException

from runpod.serverless.utils.rp_download import (
    calculate_chunk_size,
    download_files_from_urls,
    file,
)

URL_LIST = [
    "https://example.com/picture.jpg",
    "https://example.com/picture.jpg?X-Amz-Signature=123",
]

JOB_ID = "job_123"


def mock_requests_get(*args, **kwargs):
    """
    Mocks SyncClientSession.get
    """
    headers = {
        "Content-Disposition": 'attachment; filename="picture.jpg"',
        "Content-Length": "1000",
    }

    class MockResponse:
        """Mocks SyncClientSession.get response"""

        def __init__(self, content, status_code, headers=None):
            """
            Mocks SyncClientSession.get response
            """
            self.content = content
            self.status_code = status_code
            self.headers = headers or {}

        def raise_for_status(self):
            """Mocks raise_for_status function"""
            if 400 <= self.status_code < 600:
                raise RequestException(f"Status code: {self.status_code}")

        def iter_content(self, chunk_size=1024):
            """Mocks iter_content method"""
            length = len(self.content)
            for i in range(0, length, chunk_size):
                yield self.content[i : min(i + chunk_size, length)]

        def __enter__(self):
            return self

        def __exit__(self, *args):
            pass

    if args[0] in URL_LIST:
        return MockResponse(b"nothing", 200, headers)

    return MockResponse(None, 404)


class TestDownloadFilesFromUrls(unittest.TestCase):
    """Tests for download_files_from_urls"""

    def test_calculate_chunk_size(self):
        """
        Tests calculate_chunk_size
        """
        self.assertEqual(calculate_chunk_size(1024), 1024)
        self.assertEqual(calculate_chunk_size(1024 * 1024), 1024)
        self.assertEqual(calculate_chunk_size(1024 * 1024 * 1024), 1024 * 1024)
        self.assertEqual(
            calculate_chunk_size(1024 * 1024 * 1024 * 10), 1024 * 1024 * 10
        )

    @patch("os.makedirs", return_value=None)
    @patch("runpod.http_client.SyncClientSession.get", side_effect=mock_requests_get)
    @patch("builtins.open", new_callable=mock_open)
    def test_download_files_from_urls(self, mock_open_file, mock_get, mock_makedirs):
        """
        Tests download_files_from_urls
        """
        downloaded_files = download_files_from_urls(
            JOB_ID,
            [
                "https://example.com/picture.jpg",
            ],
        )

        self.assertEqual(len(downloaded_files), 1)

        # Check that the url was called with SyncClientSession.get
        self.assertIn("https://example.com/picture.jpg", mock_get.call_args_list[0][0])

        # Check that the file has the correct extension
        self.assertTrue(downloaded_files[0].endswith(".jpg"))

        mock_open_file.assert_called_once_with(downloaded_files[0], "wb")
        mock_makedirs.assert_called_once_with(
            os.path.abspath(f"jobs/{JOB_ID}/downloaded_files"), exist_ok=True
        )

        string_download_file = download_files_from_urls(
            JOB_ID, "https://example.com/picture.jpg"
        )
        self.assertTrue(string_download_file[0].endswith(".jpg"))

        # Check if None is returned when url is None
        self.assertEqual(download_files_from_urls(JOB_ID, [None]), [None])

        # Test requests exception
        mock_get.side_effect = RequestException("Error")
        self.assertEqual(
            download_files_from_urls(JOB_ID, ["https://example.com/picture.jpg"]),
            [None],
        )

    @patch("os.makedirs", return_value=None)
    @patch("runpod.http_client.SyncClientSession.get", side_effect=mock_requests_get)
    @patch("builtins.open", new_callable=mock_open)
    def test_download_files_from_urls_signed(
        self, mock_open_file, mock_get, mock_makedirs
    ):
        """
        Tests download_files_from_urls with signed urls
        """
        downloaded_files = download_files_from_urls(
            JOB_ID,
            [
                "https://example.com/picture.jpg?X-Amz-Signature=123",
            ],
        )

        # Confirms that the same number of files were downloaded as urls provided
        self.assertEqual(len(downloaded_files), 1)

        # Check that the url was called with SyncClientSession.get
        self.assertIn(URL_LIST[1], mock_get.call_args_list[0][0])

        # Check that the file has the correct extension
        self.assertTrue(downloaded_files[0].endswith(".jpg"))

        mock_open_file.assert_called_once_with(downloaded_files[0], "wb")
        mock_makedirs.assert_called_once_with(
            os.path.abspath(f"jobs/{JOB_ID}/downloaded_files"), exist_ok=True
        )


class FileDownloaderTestCase(unittest.TestCase):
    """Tests for file_downloader"""

    @patch("runpod.serverless.utils.rp_download.SyncClientSession.get")
    @patch("builtins.open", new_callable=mock_open)
    def test_download_file(self, mock_file, mock_get):
        """
        Tests download_file
        """
        # Mock the response from SyncClientSession.get
        mock_response = MagicMock()
        mock_response.content = b"file content"
        mock_response.headers = {"Content-Disposition": "filename=test_file.txt"}
        mock_get.return_value = mock_response

        # Call the function with a test URL
        result = file("http://test.com/test_file.txt")

        # Check the result
        self.assertEqual(result["type"], "txt")
        self.assertEqual(result["original_name"], "test_file.txt")
        self.assertTrue(result["file_path"].endswith(".txt"))
        self.assertIsNone(result["extracted_path"])

        # Check that the file was written correctly
        mock_file().write.assert_called_once_with(b"file content")

    @patch("runpod.serverless.utils.rp_download.SyncClientSession.get")
    @patch("builtins.open", new_callable=mock_open)
    @patch("runpod.serverless.utils.rp_download.zipfile.ZipFile")
    def test_download_zip_file(self, mock_zip, mock_file, mock_get):
        """
        Tests download_file with a zip file
        """
        # Mock the response from SyncClientSession.get
        mock_response = MagicMock()
        mock_response.content = b"zip file content"
        mock_response.headers = {"Content-Disposition": "filename=test_file.zip"}
        mock_get.return_value = mock_response

        # Call the function with a test URL
        result = file("http://test.com/test_file.zip")

        # Check the result
        self.assertEqual(result["type"], "zip")
        self.assertEqual(result["original_name"], "test_file.zip")
        self.assertTrue(result["file_path"].endswith(".zip"))
        self.assertIsNotNone(result["extracted_path"])

        # Check that the file was written correctly
        mock_file().write.assert_called_once_with(b"zip file content")

        # Check if no file name is provided
        mock_response.headers = {"Content-Disposition": ""}
        result = file("http://test.com/test_file.zip")
        self.assertEqual(result["original_name"], "test_file.zip")



""" Tests for ctl_commands.py """

import unittest
from unittest.mock import patch

from runpod.api import ctl_commands


class TestCTL(unittest.TestCase):
    """Tests for CTL Commands"""

    def test_get_user(self):
        """
        Tests get_user
        """
        with patch("runpod.api.graphql.requests.post") as patch_request:
            patch_request.return_value.json.return_value = {
                "data": {
                    "myself": {
                        "id": "USER_ID",
                    }
                }
            }

            user = ctl_commands.get_user()
            self.assertEqual(user["id"], "USER_ID")

    def test_update_user_settings(self):
        """
        Tests update_user_settings
        """
        with patch("runpod.api.graphql.requests.post") as patch_request:
            patch_request.return_value.json.return_value = {
                "data": {
                    "updateUserSettings": {"id": "USER_ID", "publicKey": "PUBLIC_KEY"}
                }
            }

            user = ctl_commands.update_user_settings("PUBLIC_KEY")
            self.assertEqual(user["id"], "USER_ID")
            self.assertEqual(user["publicKey"], "PUBLIC_KEY")

    def test_get_gpus(self):
        """
        Tests get_gpus
        """
        with patch("runpod.api.graphql.requests.post") as patch_request:
            patch_request.return_value.json.return_value = {
                "data": {
                    "gpuTypes": [
                        {
                            "id": "NVIDIA A100 80GB PCIe",
                            "displayName": "A100 80GB",
                            "memoryInGb": 80,
                        }
                    ]
                }
            }

            gpus = ctl_commands.get_gpus()

            self.assertEqual(len(gpus), 1)
            self.assertEqual(gpus[0]["id"], "NVIDIA A100 80GB PCIe")

    def test_get_gpu(self):
        """
        Tests get_gpu_by_id
        """
        with patch("runpod.api.graphql.requests.post") as patch_request:
            patch_request.return_value.json.return_value = {
                "data": {
                    "gpuTypes": [
                        {
                            "id": "NVIDIA A100 80GB PCIe",
                            "displayName": "A100 80GB",
                            "memoryInGb": 80,
                        }
                    ]
                }
            }

            gpu = ctl_commands.get_gpu("NVIDIA A100 80GB PCIe")
            self.assertEqual(gpu["id"], "NVIDIA A100 80GB PCIe")

            patch_request.return_value.json.return_value = {"data": {"gpuTypes": []}}

            with self.assertRaises(ValueError) as context:
                gpu = ctl_commands.get_gpu("Not a GPU")

            self.assertEqual(
                str(context.exception),
                "No GPU found with the specified ID, "
                "run runpod.get_gpus() to get a list of all GPUs",
            )

    def test_create_pod(self):
        """
        Tests create_pod
        """
        with patch("runpod.api.graphql.requests.post") as patch_request, patch(
            "runpod.api.ctl_commands.get_gpu"
        ) as patch_get_gpu, patch("runpod.api.ctl_commands.get_user") as patch_get_user:
            patch_request.return_value.json.return_value = {
                "data": {"podFindAndDeployOnDemand": {"id": "POD_ID"}}
            }

            patch_get_gpu.return_value = None

            patch_get_user.return_value = {
                "networkVolumes": [
                    {"id": "NETWORK_VOLUME_ID", "dataCenterId": "us-east-1"}
                ]
            }

            pod = ctl_commands.create_pod(
                name="POD_NAME",
                image_name="IMAGE_NAME",
                support_public_ip=False,
                gpu_type_id="NVIDIA A100 80GB PCIe",
                network_volume_id="NETWORK_VOLUME_ID",
            )

            self.assertEqual(pod["id"], "POD_ID")

            with self.assertRaises(ValueError) as context:
                pod = ctl_commands.create_pod(
                    name="POD_NAME",
                    cloud_type="NOT_A_CLOUD_TYPE",
                    image_name="IMAGE_NAME",
                    gpu_type_id="NVIDIA A100 80GB PCIe",
                    network_volume_id="NETWORK_VOLUME_ID",
                )

            self.assertEqual(
                str(context.exception),
                "cloud_type must be one of ALL, COMMUNITY or SECURE",
            )

    def test_stop_pod(self):
        """
        Test stop_pod
        """
        with patch("runpod.api.graphql.requests.post") as patch_request:
            patch_request.return_value.json.return_value = {
                "data": {"podStop": {"id": "POD_ID"}}
            }

            pod = ctl_commands.stop_pod(pod_id="POD_ID")

            self.assertEqual(pod["id"], "POD_ID")

    def test_resume_pod(self):
        """
        Test resume_pod
        """
        with patch("runpod.api.graphql.requests.post") as patch_request:
            patch_request.return_value.json.return_value = {
                "data": {"podResume": {"id": "POD_ID"}}
            }

            pod = ctl_commands.resume_pod(pod_id="POD_ID", gpu_count=1)

            self.assertEqual(pod["id"], "POD_ID")

    def test_terminate_pod(self):
        """
        Test terminate_pod
        """
        with patch("runpod.api.graphql.requests.post") as patch_request:
            patch_request.return_value.json.return_value = {
                "data": {"podTerminate": {"id": "POD_ID"}}
            }

            self.assertIsNone(ctl_commands.terminate_pod(pod_id="POD_ID"))

    def test_raised_error(self):
        """
        Test raised_error
        """
        with patch("runpod.api.graphql.requests.post") as patch_request:
            patch_request.return_value.json.return_value = {
                "errors": [{"message": "Error Message"}]
            }

            with self.assertRaises(Exception) as context:
                ctl_commands.get_gpus()

            self.assertEqual(str(context.exception), "Error Message")

        # Test Unauthorized with status code 401
        with patch("runpod.api.graphql.requests.post") as patch_request:
            patch_request.return_value.status_code = 401

            with self.assertRaises(Exception) as context:
                ctl_commands.get_gpus()

            self.assertEqual(
                str(context.exception),
                "Unauthorized request, please check your API key.",
            )

    def test_get_pods(self):
        """
        Tests get_pods
        """
        with patch("runpod.api.graphql.requests.post") as patch_request:
            patch_request.return_value.json.return_value = {
                "data": {
                    "myself": {
                        "pods": [
                            {
                                "id": "POD_ID",
                                "containerDiskInGb": 5,
                                "costPerHr": 0.34,
                                "desiredStatus": "RUNNING",
                                "dockerArgs": None,
                                "dockerId": None,
                                "env": [],
                                "gpuCount": 1,
                                "imageName": "runpod/pytorch:2.0.1-py3.10-cuda11.8.0-devel",
                                "lastStatusChange": "Rented by User: Tue Aug 15 2023",
                                "machineId": "MACHINE_ID",
                                "memoryInGb": 83,
                                "name": "POD_NAME",
                                "podType": "RESERVED",
                                "port": None,
                                "ports": "80/http",
                                "uptimeSeconds": 0,
                                "vcpuCount": 21,
                                "volumeInGb": 200,
                                "volumeMountPath": "/workspace",
                                "machine": {"gpuDisplayName": "RTX 3090"},
                            }
                        ]
                    }
                }
            }

            pods = ctl_commands.get_pods()

            self.assertEqual(len(pods), 1)
            self.assertEqual(pods[0]["id"], "POD_ID")

    def test_get_pod(self):
        """
        Tests get_pods
        """
        with patch("runpod.api.graphql.requests.post") as patch_request:
            patch_request.return_value.json.return_value = {
                "data": {
                    "pod": {
                        "id": "POD_ID",
                        "containerDiskInGb": 5,
                        "costPerHr": 0.34,
                        "desiredStatus": "RUNNING",
                        "dockerArgs": None,
                        "dockerId": None,
                        "env": [],
                        "gpuCount": 1,
                        "imageName": "runpod/pytorch:2.0.1-py3.10-cuda11.8.0-devel",
                        "lastStatusChange": "Rented by User: Tue Aug 15 2023",
                        "machineId": "MACHINE_ID",
                        "memoryInGb": 83,
                        "name": "POD_NAME",
                        "podType": "RESERVED",
                        "port": None,
                        "ports": "80/http",
                        "uptimeSeconds": 0,
                        "vcpuCount": 21,
                        "volumeInGb": 200,
                        "volumeMountPath": "/workspace",
                        "machine": {"gpuDisplayName": "RTX 3090"},
                    }
                }
            }

            pods = ctl_commands.get_pod("POD_ID")

            self.assertEqual(pods["id"], "POD_ID")

    def test_create_template(self):
        """
        Tests create_template
        """
        with patch("runpod.api.graphql.requests.post") as patch_request, patch(
            "runpod.api.ctl_commands.get_gpu"
        ) as patch_get_gpu:
            patch_request.return_value.json.return_value = {
                "data": {"saveTemplate": {"id": "TEMPLATE_ID"}}
            }

            patch_get_gpu.return_value = None

            template = ctl_commands.create_template(
                name="TEMPLATE_NAME", image_name="IMAGE_NAME"
            )

            self.assertEqual(template["id"], "TEMPLATE_ID")

    def test_get_endpoints(self):
        """
        Tests get_endpoints
        """
        with patch("runpod.api.graphql.requests.post") as patch_request:
            patch_request.return_value.json.return_value = {
                "data": {
                    "myself": {
                        "endpoints": [
                            {
                                "id": "ENDPOINT_ID",
                                "name": "ENDPOINT_NAME",
                                "template": {
                                    "id": "TEMPLATE_ID",
                                    "imageName": "IMAGE_NAME",
                                },
                            }
                        ]
                    }
                }
            }

            endpoints = ctl_commands.get_endpoints()

            self.assertEqual(len(endpoints), 1)
            self.assertEqual(endpoints[0]["id"], "ENDPOINT_ID")

    def test_create_endpoint(self):
        """
        Tests create_endpoint
        """
        with patch("runpod.api.graphql.requests.post") as patch_request, patch(
            "runpod.api.ctl_commands.get_gpu"
        ) as patch_get_gpu:
            patch_request.return_value.json.return_value = {
                "data": {"saveEndpoint": {"id": "ENDPOINT_ID"}}
            }

            patch_get_gpu.return_value = None

            endpoint = ctl_commands.create_endpoint(
                name="ENDPOINT_NAME", template_id="TEMPLATE_ID"
            )

            self.assertEqual(endpoint["id"], "ENDPOINT_ID")

    def test_update_endpoint_template(self):
        """
        Tests update_endpoint_template
        """
        with patch("runpod.api.graphql.requests.post") as patch_request, patch(
            "runpod.api.ctl_commands.get_gpu"
        ) as patch_get_gpu:
            patch_request.return_value.json.return_value = {
                "data": {"updateEndpointTemplate": {"id": "ENDPOINT_ID"}}
            }

            patch_get_gpu.return_value = None

            endpoint = ctl_commands.update_endpoint_template(
                endpoint_id="ENDPOINT_ID", template_id="TEMPLATE_ID"
            )

            self.assertEqual(endpoint["id"], "ENDPOINT_ID")

    @patch("runpod.api.ctl_commands.run_graphql_query")
    def test_create_container_registry_auth(self, mock_run_graphql_query):
        """
        Tests create_container_registry_auth by mocking the run_graphql_query function
        """
        # Setup the mock to return a predefined response
        mock_run_graphql_query.return_value = {
            "data": {
                "saveRegistryAuth": {"id": "REGISTRY_AUTH_ID", "name": "REGISTRY_NAME"}
            }
        }

        # Call the function under test with dummy arguments
        result = ctl_commands.create_container_registry_auth(
            name="REGISTRY_NAME", username="username", password="password"
        )

        # Assertions to verify the function behavior
        self.assertEqual(result["id"], "REGISTRY_AUTH_ID")
        self.assertEqual(result["name"], "REGISTRY_NAME")

        # Verify that run_graphql_query was called with the correct parameters
        mock_run_graphql_query.assert_called_once()  # Ensure it was called exactly once

        # Access the first (and only) call's arguments directly
        called_args, _ = mock_run_graphql_query.call_args

        # The GraphQL query is expected to be the first positional argument in the call
        graphql_query = called_args[0]

        self.assertIn("mutation SaveRegistryAuth", graphql_query)
        self.assertIn("REGISTRY_NAME", graphql_query)
        self.assertIn("username", graphql_query)
        self.assertIn("password", graphql_query)



"""
RunPod | CLI | Entry

The entry point for the CLI.
"""

import click

from .groups.config.commands import config_wizard
from .groups.exec.commands import exec_cli
from .groups.pod.commands import pod_cli
from .groups.project.commands import project_cli
from .groups.ssh.commands import ssh_cli


@click.group()
def runpod_cli():
    """A collection of CLI functions for RunPod."""


runpod_cli.add_command(config_wizard)  # runpod config

runpod_cli.add_command(ssh_cli)  # runpod ssh
runpod_cli.add_command(pod_cli)  # runpod pod
runpod_cli.add_command(exec_cli)  # runpod exec
runpod_cli.add_command(project_cli)  # runpod project



"""
RunPod | API | Mutations | User
"""


def generate_user_mutation(pubkey):
    """'
    Generates a mutation to edit a user.
    """
    input_fields = []

    escaped_pubkey = pubkey.replace("\n", "\\n")
    input_fields.append(f'pubKey: "{escaped_pubkey}"')

    # Format input fields
    input_string = ", ".join(input_fields)

    return f"""
    mutation {{
        updateUserSettings(
            input: {{
                {input_string}
            }}
        ) {{
            id
            pubKey
        }}
    }}
    """



"""
runpod | serverless | rp_debugger.py

A collection of functions to help with debugging.
"""

import datetime
import platform
import time

import cpuinfo

# ---------------------------------------------------------------------------- #
#                                  System Info                                 #
# ---------------------------------------------------------------------------- #
OS_INFO = f"{platform.system()} {platform.release()}"

try:
    PROCESSOR = cpuinfo.get_cpu_info()["brand_raw"]
except KeyError:
    PROCESSOR = "Unable to get processor info."

PYTHON_VERSION = platform.python_version()


class Checkpoints:
    """
    A singleton class to store checkpoint times.

    Format:
    [
        {
            'name': 'checkpoint_name',
            'start': 1234567890.123456,
            'end': 1234567890.123456
            'duration_ms': 1234567890.123456
        },
    ]

    Usage:
    from rp_debugger import Checkpoints

    checkpoints = Checkpoints()

    # Add a checkpoint
    checkpoints.add('checkpoint_name')

    # Start a checkpoint
    checkpoints.start('checkpoint_name')

    # Stop a checkpoint
    checkpoints.stop('checkpoint_name')
    """

    __instance = None
    checkpoints = []
    name_lookup = {}

    def __new__(cls):
        if Checkpoints.__instance is None:
            Checkpoints.__instance = object.__new__(cls)
            Checkpoints.__instance.checkpoints = []
            Checkpoints.__instance.name_lookup = {}
        return Checkpoints.__instance

    def add(self, name):
        """
        Add a checkpoint.
        Returns the index of the checkpoint.
        """
        if name in self.name_lookup:
            raise KeyError(f'Checkpoint name "{name}" already exists.')

        self.checkpoints.append({"name": name})

        index = len(self.checkpoints) - 1
        self.name_lookup[name] = index

    def start(self, name):
        """
        Start a checkpoint.
        """
        if name not in self.name_lookup:
            raise KeyError(f"Checkpoint name '{name}' does not exist.")

        index = self.name_lookup[name]
        self.checkpoints[index]["start"] = time.perf_counter()
        self.checkpoints[index]["start_utc"] = (
            datetime.datetime.utcnow().isoformat() + "Z"
        )

    def stop(self, name):
        """
        Stop a checkpoint.
        """
        if name not in self.name_lookup:
            raise KeyError(f"Checkpoint name '{name}' does not exist.")

        index = self.name_lookup[name]

        if "start" not in self.checkpoints[index]:
            raise KeyError("Checkpoint has not been started.")

        self.checkpoints[index]["end"] = time.perf_counter()
        self.checkpoints[index]["stop_utc"] = (
            datetime.datetime.utcnow().isoformat() + "Z"
        )

    def get_checkpoints(self):
        """
        Get the results of the checkpoints.
        """
        results = []
        for checkpoint in self.checkpoints:
            if "start" not in checkpoint or "end" not in checkpoint:
                continue
            start_time = checkpoint["start"]
            end_time = checkpoint["end"]
            checkpoint["duration_ms"] = (end_time - start_time) * 1000

            checkpoint.pop("start")
            checkpoint.pop("end")

            results.append(checkpoint)

        return results

    def clear(self):
        """
        Clear the checkpoints.
        """
        self.checkpoints = []
        self.name_lookup = {}


class LineTimer:
    """
    A utility that can be used to time code execution using the with statement.
    When used the times should be added to the checkpoints object.
    """

    def __init__(self, name):
        self.checkpoints = Checkpoints()
        self.name = name
        self.checkpoints.add(self.name)

    def __enter__(self):
        self.checkpoints.start(self.name)

    def __exit__(self, *args):
        self.checkpoints.stop(self.name)


class FunctionTimer:  # pylint: disable=too-few-public-methods
    """
    A class-based decorator to benchmark a function.
    """

    def __init__(self, function):
        self.function = function
        self.checkpoints = Checkpoints()

    def __call__(self, *args, **kwargs):
        self.checkpoints.add(self.function.__name__)

        try:
            self.checkpoints.start(self.function.__name__)
            result = self.function(*args, **kwargs)

        finally:
            if self.function.__name__ in self.checkpoints.name_lookup:
                self.checkpoints.stop(self.function.__name__)

        return result


def get_debugger_output():
    """
    Return the debugger output.
    """
    print("Getting debugger output...")
    import runpod  # pylint: disable=import-outside-toplevel, cyclic-import

    print("Getting checkpoints...")

    checkpoints = Checkpoints()
    ckpt_results = checkpoints.get_checkpoints()
    checkpoints.clear()

    print("Getting system info...")

    system_info = {
        "os": OS_INFO,
        "processor": PROCESSOR,
        "python_version": PYTHON_VERSION,
        "runpod": runpod.__version__,
    }

    print("Debugger output complete.")

    return {
        "system_info": system_info,
        "timestamps": ckpt_results,
    }


def clear_debugger_output():
    """
    Clear the debugger output.
    """
    checkpoints = Checkpoints()
    checkpoints.clear()



"""
RunPod | CLI | SSH | Commands
"""

import click
from prettytable import PrettyTable

from .functions import generate_ssh_key_pair, get_user_pub_keys


@click.group("ssh", help="Manage and configure SSH keys for secure access to pods.")
def ssh_cli():
    """Manage and configure SSH keys."""


@ssh_cli.command("list-keys")
def list_keys():
    """
    Lists the SSH keys for the current user.
    """
    key_list = get_user_pub_keys()
    table = PrettyTable(["Public Key", "Type", "Fingerprint"])
    for key in key_list:
        table.add_row((key["name"], key["type"], key["fingerprint"]))
    click.echo(table)


@ssh_cli.command("add-key")
@click.option("--key", default=None, help="The public key to add.")
@click.option(
    "--key-file", default=None, help="The file containing the public key to add."
)
def add_key(key, key_file):
    """
    Adds an SSH key to the current user account.
    If no key is provided, one will be generated.
    """
    if not key and not key_file:
        click.confirm("Would you like to add an SSH key to your account?", abort=True)
        key_name = click.prompt(
            "Please enter a name for this key", default="RunPod-Key", type=str
        )
        key_name = key_name.replace(" ", "-")
        generate_ssh_key_pair(key_name)

    click.echo("The key has been added to your account.")



""" Test API Wrapper Pod Mutations """

import unittest

from runpod.api.mutations import pods


class TestPodMutations(unittest.TestCase):
    """Test API Wrapper Pod Mutations"""

    def test_generate_pod_deployment_mutation(self):
        """
        Test generate_pod_deployment_mutation
        """
        result = pods.generate_pod_deployment_mutation(
            name="test",
            image_name="test_image",
            gpu_type_id="1",
            cloud_type="cloud",
            data_center_id="1",
            country_code="US",
            gpu_count=1,
            volume_in_gb=100,
            container_disk_in_gb=10,
            min_vcpu_count=1,
            min_memory_in_gb=1,
            docker_args="args",
            ports="8080",
            volume_mount_path="/path",
            env={"ENV": "test"},
            support_public_ip=True,
            template_id="abcde",
            allowed_cuda_versions=["11.8", "12.0"],
        )

        # Here you should check the correct structure of the result
        self.assertIn("mutation", result)

    def test_generate_pod_stop_mutation(self):
        """
        Test generate_pod_stop_mutation
        """
        result = pods.generate_pod_stop_mutation("pod_id")
        # Here you should check the correct structure of the result
        self.assertIn("mutation", result)

    def test_generate_pod_resume_mutation(self):
        """
        Test generate_pod_resume_mutation
        """
        result = pods.generate_pod_resume_mutation("pod_id", 1)
        # Here you should check the correct structure of the result
        self.assertIn("mutation", result)

    def test_generate_pod_terminate_mutation(self):
        """
        Test generate_pod_terminate_mutation
        """
        result = pods.generate_pod_terminate_mutation("pod_id")
        # Here you should check the correct structure of the result
        self.assertIn("mutation", result)


if __name__ == "__main__":
    unittest.main()



""" Unit tests for the function generate_pod_template in the file api_wrapper.py """

import unittest

from runpod.api.mutations.templates import generate_pod_template


class TestGeneratePodTemplate(unittest.TestCase):
    """Unit tests for the function generate_pod_template in the file api_wrapper.py"""

    def test_basic_required_fields(self):
        """Test the basic required fields are present in the generated template"""
        result = generate_pod_template("test_name", "test_image_name")
        self.assertIn('name: "test_name"', result)
        self.assertIn('imageName: "test_image_name"', result)
        self.assertIn('dockerArgs: ""', result)  # Defaults
        self.assertIn("containerDiskInGb: 10", result)  # Defaults
        self.assertIn("volumeInGb: 0", result)  # Defaults
        self.assertIn('ports: ""', result)  # Defaults
        self.assertIn("env: []", result)  # Defaults
        self.assertIn("isServerless: false", result)  # Defaults

    def test_optional_fields(self):
        """Test the optional fields are present in the generated template"""
        result = generate_pod_template(
            "test_name",
            "test_image_name",
            docker_start_cmd="test_cmd",
            volume_in_gb=5,
            volume_mount_path="/path/to/volume",
            ports="8000, 8001",
            env={"VAR1": "val1", "VAR2": "val2"},
            is_serverless=True,
            registry_auth_id="test_auth",
        )
        self.assertIn('dockerArgs: "test_cmd"', result)
        self.assertIn("volumeInGb: 5", result)
        self.assertIn('volumeMountPath: "/path/to/volume"', result)
        self.assertIn('ports: "8000,8001"', result)
        self.assertIn(
            'env: [{ key: "VAR1", value: "val1" }, { key: "VAR2", value: "val2" }]',
            result,
        )
        self.assertIn("isServerless: true", result)
        self.assertIn('containerRegistryAuthId : "test_auth"', result)



"""Tests for the SSH commands of the CLI."""

import unittest
from unittest.mock import patch

from click.testing import CliRunner

from runpod.cli.groups.ssh.commands import add_key, list_keys


class TestSSHCommands(unittest.TestCase):
    """Tests for the SSH commands of the CLI."""

    def test_list_keys(self):
        """Test the list_keys command."""
        runner = CliRunner()
        with patch(
            "runpod.cli.groups.ssh.commands.get_user_pub_keys",
            return_value=[
                {"name": "key1", "type": "RSA", "fingerprint": "fp1"},
                {"name": "key2", "type": "DSA", "fingerprint": "fp2"},
            ],
        ) as mock_get_keys:

            result = runner.invoke(list_keys)

            self.assertIn("key1", result.output)
            self.assertIn("key2", result.output)

            assert mock_get_keys.called

    def test_add_key_without_params(self):
        """Test the add_key command without parameters."""
        runner = CliRunner()
        with patch(
            "runpod.cli.groups.ssh.commands.generate_ssh_key_pair"
        ) as mock_gen_key, patch(
            "runpod.cli.groups.ssh.commands.click.prompt", return_value="TestKey"
        ) as mock_prompt, patch(
            "runpod.cli.groups.ssh.commands.click.confirm", return_value=True
        ) as mock_confirm:  # pylint: disable=line-too-long

            result = runner.invoke(add_key, [])

            self.assertIn("The key has been added to your account.", result.output)
            mock_gen_key.assert_called_once_with("TestKey")
            assert mock_prompt.called
            assert mock_confirm.called



"""
Tests for the rp_progress.py module.
"""

import unittest
from threading import Event
from unittest.mock import ANY, patch

from runpod.serverless.modules.rp_progress import _thread_target, progress_update


class TestProgressUpdate(unittest.TestCase):
    """Tests for the progress_update function."""

    @patch("runpod.serverless.modules.rp_progress.send_result")
    @patch("runpod.serverless.modules.rp_progress._thread_target")
    def test_progress_update(self, mock_thread_target, mock_result):
        """
        Tests that the progress_update function.
        """
        # Create an event to track thread completion
        thread_event = Event()

        def mock_thread_function(job, progress):
            try:
                assert job == "fake_job", "Job ID was not passed correctly"
                assert progress == "50%", "Progress was not passed correctly"
            except Exception as err:  # pylint: disable=broad-except
                print(f"Exception in mocked function: {err}")
            finally:
                thread_event.set()

        mock_thread_target.side_effect = mock_thread_function

        # Call the function
        job = {"id": "fake_job"}
        progress = "50%"
        progress_update(job, progress)
        _thread_target(job, progress)

        assert mock_thread_target.called, "Thread function was not started"
        mock_thread_target.assert_called_once_with(job, progress)
        assert thread_event.wait(
            timeout=30
        ), "Thread did not complete within expected time"

        # Assertions
        expected_job_data = {"status": "IN_PROGRESS", "output": progress}
        mock_result.assert_called_once_with(ANY, expected_job_data, job)



"""Unit tests for the error classes in the runpod.error module."""

import unittest

# Assuming the error classes are in a file named 'error.py'
from runpod.error import AuthenticationError, QueryError, RunPodError


class TestErrorClasses(unittest.TestCase):
    """Unit tests for the error classes in the runpod.error module."""

    def test_run_pod_error_with_message(self):
        """Test the RunPodError class with a message."""
        error_msg = "An error occurred"
        err = RunPodError(error_msg)
        self.assertEqual(str(err), error_msg)

    def test_run_pod_error_without_message(self):
        """Test the RunPodError class without a message."""
        err = RunPodError()
        self.assertEqual(str(err), "None")

    def test_authentication_error(self):
        """Test the AuthenticationError class."""
        error_msg = "Authentication failed"
        err = AuthenticationError(error_msg)
        self.assertEqual(str(err), error_msg)

    def test_query_error_with_message_and_query(self):
        """Test the QueryError class with a message and query."""
        error_msg = "Query failed"
        query_str = "SELECT * FROM some_table WHERE condition"
        err = QueryError(error_msg, query_str)
        self.assertEqual(str(err), error_msg)
        self.assertEqual(err.query, query_str)

    def test_query_error_without_message_and_query(self):
        """Test the QueryError class without a message or query."""
        err = QueryError()
        self.assertEqual(str(err), "None")
        self.assertIsNone(err.query)



""" Core functionality for the runpod serverless worker. """

import asyncio
import ctypes
import inspect
import json
import os
import pathlib
import typing
from ctypes import CDLL, byref, c_char_p, c_int
from typing import Any, Callable, Dict, List, Optional

from runpod.serverless.modules import rp_job
from runpod.serverless.modules.rp_logger import RunPodLogger
from runpod.version import __version__ as runpod_version

log = RunPodLogger()

# _runpod_sls_get_jobs status codes
STILL_WAITING = 0 
OK = 1
ERROR_FROM_SERVER = 2
ERROR_BUFFER_TOO_SMALL = 3

class CGetJobResult(ctypes.Structure):  # pylint: disable=too-few-public-methods
    """
     result of _runpod_sls_get_jobs.
    ## fields
    - `res_len` the number bytes were written to the `dst_buf` passed to _runpod_sls_get_jobs.
    - `status_code` tells you what happened.
    see CGetJobResult.status_code for more information.
    """

    _fields_ = [("status_code", ctypes.c_int), ("res_len", ctypes.c_int)]

    def __str__(self) -> str:
        return f"CGetJobResult(res_len={self.res_len}, status_code={self.status_code})"


def notregistered():
    """Function to raise NotImplementedError"""
    raise RuntimeError("This function is not registered with the SLS Core.")

class SlsCoreError(Exception):
    pass

class Hook:  # pylint: disable=too-many-instance-attributes
    """Singleton class for interacting with sls_core.so"""

    _instance = None

    # C function pointers
    _get_jobs: Callable = notregistered
    _progress_update: Callable = notregistered
    _stream_output: Callable = notregistered
    _post_output: Callable = notregistered
    _finish_stream: Callable = notregistered

    def __new__(cls):
        if Hook._instance is None:
            log.debug("SLS Core | Initializing Hook.")
            Hook._instance = object.__new__(cls)
            Hook._initialized = False

        return Hook._instance

    def __init__(self, rust_so_path: Optional[str] = None) -> None:

        if self._initialized:
            return

        if rust_so_path is None:
            default_path = os.path.join(
                pathlib.Path(__file__).parent.absolute(), "sls_core.so"
            )
            self.rust_so_path = os.environ.get(
                "RUNPOD_SLS_CORE_PATH", str(default_path)
            )
        else:
            self.rust_so_path = rust_so_path

        rust_library = CDLL(self.rust_so_path)
        buffer = ctypes.create_string_buffer(1024)  # 1 KiB
        num_bytes = rust_library._runpod_sls_crate_version(
            byref(buffer), c_int(len(buffer))
        )

        self.rust_crate_version = buffer.raw[:num_bytes].decode("utf-8")

        # Get Jobs
        self._get_jobs = rust_library._runpod_sls_get_jobs
        self._get_jobs.restype = CGetJobResult

        # Progress Update
        self._progress_update = rust_library._runpod_sls_progress_update
        self._progress_update.argtypes = [
            c_char_p,
            c_int,  # id_ptr, id_len
            c_char_p,
            c_int,  # json_ptr, json_len
        ]
        self._progress_update.restype = c_int  # 1 if success, 0 if failure

        # Stream Output
        self._stream_output = rust_library._runpod_sls_stream_output
        self._stream_output.argtypes = [
            c_char_p,
            c_int,  # id_ptr, id_len
            c_char_p,
            c_int,  # json_ptr, json_len
        ]
        self._stream_output.restype = c_int  # 1 if success, 0 if failure

        # Post Output
        self._post_output = rust_library._runpod_sls_post_output
        self._post_output.argtypes = [
            c_char_p,
            c_int,  # id_ptr, id_len
            c_char_p,
            c_int,  # json_ptr, json_len
        ]
        self._post_output.restype = c_int  # 1 if success, 0 if failure

        # Finish Stream
        self._finish_stream = rust_library._runpod_sls_finish_stream
        self._finish_stream.argtypes = [c_char_p, c_int]  # id_ptr, id_len
        self._finish_stream.restype = c_int  # 1 if success, 0 if failure

        rust_library._runpod_sls_crate_version.restype = c_int

        rust_library._runpod_sls_init.argtypes = []
        rust_library._runpod_sls_init.restype = c_int
        rust_library._runpod_sls_init()

        self._initialized = True

    def _json_serialize_job_data(self, job_data: Any) -> bytes:
        return json.dumps(job_data, ensure_ascii=False).encode("utf-8")

    def get_jobs(self, max_concurrency: int, max_jobs: int) -> List[Dict[str, Any]]:
        """Get a job or jobs from the queue. The jobs are returned as a list of Job objects."""
        buf = ctypes.create_string_buffer(
            1024 * 1024 * 20
        )  # 20MB buffer to store jobs in
        destination_length = len(buf.raw)
        res: CGetJobResult = self._get_jobs(
            c_int(max_concurrency),
            c_int(max_jobs),
            byref(buf),
            c_int(destination_length),
        )
        n = res.res_len
        code = res.status_code
        if code == STILL_WAITING:
            return []  # still waiting for jobs
        elif code == OK:  # success! the job was stored bytes 0..res_len of buf.raw
            log.trace(f"decoding {n} bytes of JSON")
            return list(json.loads(buf.raw[:n].decode("utf-8")))
        elif code == ERROR_FROM_SERVER:
            try:
                b = buf.raw[: res.res_len].decode("utf-8")
            except Exception:
                b = "<failed to decode buffer>"
            if b == "":
                b = "<unknown error or buffer too small>"
            raise SlsCoreError(f"_runpod_sls_get_jobs: status code 2: error from server: {b}")
        elif code == ERROR_BUFFER_TOO_SMALL:  # buffer too small
            raise SlsCoreError("_runpod_sls_get_jobs: status code 3: buffer too small")
        else:
            raise ValueError(f"_runpod_sls_get_jobs: unknown status code {code}")


    def progress_update(self, job_id: str, json_data: bytes) -> bool:
        """
        send a progress update to AI-API.
        """
        id_bytes = job_id.encode("utf-8")
        return bool(
            self._progress_update(
                c_char_p(id_bytes),
                c_int(len(id_bytes)),
                c_char_p(json_data),
                c_int(len(json_data)),
            )
        )

    async def stream_output(self, job_id: str, job_output: bytes) -> bool:
        """
        send part of a streaming result to AI-API.
        """
        json_data = self._json_serialize_job_data(job_output)
        id_bytes = job_id.encode("utf-8")
        return bool(
            self._stream_output(
                c_char_p(id_bytes),
                c_int(len(id_bytes)),
                c_char_p(json_data),
                c_int(len(json_data)),
            )
        )

    def post_output(self, job_id: str, job_output: bytes) -> bool:
        """
        send the result of a job to AI-API.
        Returns True if the task was successfully stored, False otherwise.
        """
        json_data = self._json_serialize_job_data(job_output)
        id_bytes = job_id.encode("utf-8")
        return bool(
            self._post_output(
                c_char_p(id_bytes),
                c_int(len(id_bytes)),
                c_char_p(json_data),
                c_int(len(json_data)),
            )
        )

    def finish_stream(self, job_id: str) -> bool:
        """
        tell the SLS queue that the result of a streaming job is complete.
        """
        id_bytes = job_id.encode("utf-8")
        return bool(self._finish_stream(c_char_p(id_bytes), c_int(len(id_bytes))))


# -------------------------------- Process Job ------------------------------- #
async def _process_job(
    config: Dict[str, Any], job: Dict[str, Any], hook
) -> Dict[str, Any]:
    """Process a single job."""
    handler = config["handler"]

    result = {}
    try:
        if inspect.isgeneratorfunction(handler) or inspect.isasyncgenfunction(handler):
            log.debug("SLS Core | Running job as a generator.")
            generator_output = rp_job.run_job_generator(handler, job)
            aggregated_output: dict[str, typing.Any] = {"output": []}

            async for part in generator_output:
                log.trace(f"SLS Core | Streaming output: {part}", job["id"])

                if "error" in part:
                    aggregated_output = part
                    break
                if config.get("return_aggregate_stream", False):
                    aggregated_output["output"].append(part["output"])

                await hook.stream_output(job["id"], part)

            log.debug("SLS Core | Finished streaming output.", job["id"])
            hook.finish_stream(job["id"])
            result = aggregated_output

        else:
            log.debug("SLS Core | Running job as a standard function.")
            result = await rp_job.run_job(handler, job)
            result = result.get("output", result)

    except Exception as err:  # pylint: disable=broad-except
        log.error(f"SLS Core | Error running job: {err}", job["id"])
        result = {"error": str(err)}

    finally:
        log.debug(f"SLS Core | Posting output: {result}", job["id"])
        hook.post_output(job["id"], result)
        return result


# ---------------------------------------------------------------------------- #
#                                  Run Worker                                  #
# ---------------------------------------------------------------------------- #
async def run(config: Dict[str, Any]) -> None:
    """Run the worker.

    Args:
        config: A dictionary containing the following keys:
            handler: A function that takes a job and returns a result.
    """
    max_concurrency = config.get("max_concurrency", 1)
    max_jobs = config.get("max_jobs", 1)

    serverless_hook = Hook()
    while True:
        try:
            jobs = serverless_hook.get_jobs(max_concurrency, max_jobs)
        except SlsCoreError as err:
            log.error(f"SLS Core | Error getting jobs: {err}")
            await asyncio.sleep(0.2) # sleep for a bit before trying again
            continue

        if len(jobs) == 0 or jobs is None:
            await asyncio.sleep(0)
            continue

        for job in jobs:
            asyncio.create_task(
                _process_job(config, job, serverless_hook), name=job["id"]
            )
            await asyncio.sleep(0)

        await asyncio.sleep(0)


def main(config: Dict[str, Any]) -> None:
    """Run the worker in an asyncio event loop."""
    if config.get("handler") is None:
        log.error("SLS Core | config must contain a handler function")
        raise ValueError("config must contain a handler function")

    try:
        work_loop = asyncio.new_event_loop()
        asyncio.ensure_future(run(config), loop=work_loop)
        work_loop.run_forever()
    finally:
        work_loop.close()



""" Allows endpoints to be imported as a module. """

from .asyncio.asyncio_runner import Endpoint as AsyncioEndpoint
from .asyncio.asyncio_runner import Job as AsyncioJob
from .runner import Endpoint, Job



"""
runpod-python | setup.py
Called to setup the runpod-python package.
"""

from setuptools import find_packages, setup

# README.md > long_description
with open("README.md", encoding="utf-8") as long_description_file:
    long_description = long_description_file.read()

# requirements.txt > requirements
with open("requirements.txt", encoding="UTF-8") as requirements_file:
    install_requires = requirements_file.read().splitlines()

extras_require = {
    "test": [
        "asynctest",
        "faker",
        "nest_asyncio",
        "pytest",
        "pytest-cov",
        "pytest-timeout",
        "pytest-asyncio",
    ]
}

if __name__ == "__main__":

    setup(
        name="runpod",
        use_scm_version=True,
        setup_requires=["setuptools>=45", "setuptools_scm", "wheel"],
        install_requires=install_requires,
        extras_require=extras_require,
        packages=find_packages(),
        python_requires=">=3.8",
        description="🐍 | Python library for RunPod API and serverless worker SDK.",
        long_description=long_description,
        long_description_content_type="text/markdown",
        author="RunPod",
        author_email="engineer@runpod.io",
        url="https://runpod.io",
        project_urls={
            "Documentation": "https://docs.runpod.io",
            "Source": "https://github.com/runpod/runpod-python",
            "Bug Tracker": "https://github.com/runpod/runpod-python/issues",
            "Changelog": "https://github.com/runpod/runpod-python/blob/main/CHANGELOG.md",
        },
        classifiers=[
            "Environment :: Web Environment",
            "Intended Audience :: Developers",
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
            "Programming Language :: Python",
            "Programming Language :: Python :: 3",
            "Topic :: Internet :: WWW/HTTP",
            "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
        ],
        include_package_data=True,
        entry_points={"console_scripts": ["runpod = runpod.cli.entry:runpod_cli"]},
        keywords=[
            "runpod",
            "ai",
            "gpu",
            "serverless",
            "SDK",
            "API",
            "python",
            "library",
        ],
        license="MIT",
    )



"""Utility function for transforming HuggingFace repositories into model-cache paths"""

import typing
from runpod.serverless.modules.rp_logger import RunPodLogger

log = RunPodLogger()


def resolve_model_cache_path_from_hugginface_repository(
    huggingface_repository: str,
    /,
    path_template: str = "/runpod/cache/{model}/{revision}",  # TODO: Should we just hardcode this?
) -> typing.Union[str, None]:
    """
    Resolves the model-cache path for a HuggingFace model based on its repository string.

    Args:
        huggingface_repository (str): Repository string in format "model_name:revision" or
                                    "org/model_name:revision". If no revision is specified,
                                    "main" is used. For example:
                                    - "runwayml/stable-diffusion-v1-5:experimental"
                                    - "runwayml/stable-diffusion-v1-5" (uses "main" revision)
                                    - "stable-diffusion-v1-5:main"
        path_template (str, optional): Template string for the cache path. Must contain {model}
                                     and {revision} placeholders. Defaults to "/runpod/cache/{model}/{revision}".

    Returns:
        str | None: Absolute path where the model is cached, following the template provided in path_template. Returns None if no model name could be extracted.

    Examples:
        >>> resolve_model_cache_path_from_hugginface_repository("runwayml/stable-diffusion-v1-5:experimental")
        "/runpod/cache/runwayml/stable-diffusion-v1-5/experimental"
        >>> resolve_model_cache_path_from_hugginface_repository("runwayml/stable-diffusion-v1-5")
        "/runpod/cache/runwayml/stable-diffusion-v1-5/main"
        >>> resolve_model_cache_path_from_hugginface_repository(":experimental")
        None
    """
    model, *revision = huggingface_repository.rsplit(":", 1)
    if not model:
        # We could throw an exception here but returning None allows us to filter a list of repositories without needing a try/except block
        log.warn(  # type: ignore in strict mode the typechecker complains about this method being partially unknown
            f'Unable to resolve the model-cache path for "{huggingface_repository}"'
        )
        return None
    return path_template.format(
        model=model, revision=revision[0] if revision else "main"
    )



""" Test CLI pod commands """

import unittest
from unittest.mock import MagicMock, patch

from click.testing import CliRunner
from prettytable import PrettyTable

from runpod.cli.entry import runpod_cli


class TestPodCommands(unittest.TestCase):
    """Test CLI pod commands"""

    @patch("runpod.cli.groups.pod.commands.get_pods")
    @patch("runpod.cli.groups.pod.commands.click.echo")
    def test_list_pods(self, mock_echo, mock_get_pods):
        """
        Test list_pods
        """
        # Mock data returned by get_pods
        mock_get_pods.return_value = [
            {
                "id": "1",
                "name": "Pod1",
                "desiredStatus": "Running",
                "imageName": "Image1",
            },
            {
                "id": "2",
                "name": "Pod2",
                "desiredStatus": "Stopped",
                "imageName": "Image2",
            },
        ]

        runner = CliRunner()
        result = runner.invoke(runpod_cli, ["pod", "list"])

        # Create expected table
        assert result.exit_code == 0, result.exception
        expected_table = PrettyTable(["ID", "Name", "Status", "Image"])
        expected_table.add_row(("1", "Pod1", "Running", "Image1"))
        expected_table.add_row(("2", "Pod2", "Stopped", "Image2"))

        # Assert that click.echo was called with the correct table
        mock_echo.assert_called()

    @patch("runpod.cli.groups.pod.commands.click.prompt")
    @patch("runpod.cli.groups.pod.commands.click.confirm")
    @patch("runpod.cli.groups.pod.commands.click.echo")
    @patch("runpod.cli.groups.pod.commands.create_pod")
    def test_create_new_pod(
        self, mock_create_pod, mock_echo, mock_confirm, mock_prompt
    ):  # pylint: disable=too-many-arguments,line-too-long
        """
        Test create_new_pod
        """
        # Mock values
        mock_confirm.return_value = True  # for the quick_launch option
        mock_prompt.return_value = "RunPod-CLI-Pod"
        mock_create_pod.return_value = {"id": "sample_id"}

        runner = CliRunner()
        result = runner.invoke(runpod_cli, ["pod", "create"])

        # Assertions
        assert result.exit_code == 0, result.exception
        mock_prompt.assert_called_once_with("Enter pod name", default="RunPod-CLI-Pod")
        mock_echo.assert_called_with("Pod sample_id has been created.")
        mock_create_pod.assert_called_with(
            "RunPod-CLI-Pod",
            "runpod/base:0.0.0",
            "NVIDIA GeForce RTX 3090",
            gpu_count=1,
            support_public_ip=True,
            ports="22/tcp",
        )
        mock_echo.assert_called_with("Pod sample_id has been created.")

    @patch("runpod.cli.groups.pod.commands.click.echo")
    @patch("runpod.cli.groups.pod.commands.ssh_cmd.SSHConnection")
    def test_connect_to_pod(self, mock_ssh_connection, mock_echo):
        """
        Test connect_to_pod function
        """
        pod_id = "sample_id"

        mock_ssh = MagicMock()
        mock_ssh_connection.return_value = mock_ssh

        runner = CliRunner()
        result = runner.invoke(runpod_cli, ["pod", "connect", pod_id])

        assert result.exit_code == 0, result.exception
        mock_echo.assert_called_once_with(f"Connecting to pod {pod_id}...")
        mock_ssh_connection.assert_called_once_with(pod_id)
        mock_ssh.launch_terminal.assert_called_once_with()



"""
RunPod | CLI | Pod | Commands
"""

import click
from prettytable import PrettyTable

from runpod import create_pod, get_pods

from ...utils import ssh_cmd


@click.group("pod", help="Manage and interact with pods.")
def pod_cli():
    """A collection of CLI functions for Pod."""


@pod_cli.command("list")
def list_pods():
    """
    Lists the pods for the current user.
    """
    table = PrettyTable(["ID", "Name", "Status", "Image"])
    for pod in get_pods():
        table.add_row((pod["id"], pod["name"], pod["desiredStatus"], pod["imageName"]))

    click.echo(table)


@pod_cli.command("create")
@click.argument("name", required=False)
@click.option("--image", default=None, help="The image to use for the pod.")
@click.option("--gpu-type", default=None, help="The GPU type to use for the pod.")
@click.option("--gpu-count", default=1, help="The number of GPUs to use for the pod.")
@click.option(
    "--support-public-ip", default=True, help="Whether or not to support a public IP."
)
def create_new_pod(
    name, image, gpu_type, gpu_count, support_public_ip
):  # pylint: disable=too-many-arguments
    """
    Creates a pod.
    """
    kwargs = {
        "gpu_count": gpu_count,
        "support_public_ip": support_public_ip,
    }

    if not name:
        name = click.prompt("Enter pod name", default="RunPod-CLI-Pod")

    quick_launch = click.confirm("Would you like to launch default pod?", abort=True)
    if quick_launch:
        image = "runpod/base:0.0.0"
        gpu_type = "NVIDIA GeForce RTX 3090"
        kwargs["ports"] = "22/tcp"

        click.echo("Launching default pod...")

    new_pod = create_pod(name, image, gpu_type, **kwargs)

    click.echo(f'Pod {new_pod["id"]} has been created.')


@pod_cli.command("connect")
@click.argument("pod_id")
def connect_to_pod(pod_id):
    """
    Connects to a pod.
    """
    click.echo(f"Connecting to pod {pod_id}...")
    ssh = ssh_cmd.SSHConnection(pod_id)
    ssh.launch_terminal()



"""
Unit tests for the config command.
"""

import unittest
from unittest.mock import mock_open, patch

from runpod.cli.groups.config import functions


class TestConfig(unittest.TestCase):
    """Unit tests for the config function."""

    def setUp(self) -> None:
        self.sample_credentials = "[default]\n" 'api_key = "RUNPOD_API_KEY"\n'

    @patch("runpod.cli.groups.config.functions.toml.load")
    @patch("builtins.open", new_callable=mock_open())
    def test_set_credentials(self, mock_file, mock_toml_load):
        """
        Tests the set_credentials function.
        """
        mock_toml_load.return_value = ""
        functions.set_credentials("RUNPOD_API_KEY")

        mock_file.assert_called_with(functions.CREDENTIAL_FILE, "w", encoding="UTF-8")

        with self.assertRaises(ValueError) as context:
            mock_toml_load.return_value = {"default": True}
            functions.set_credentials("RUNPOD_API_KEY")

        self.assertEqual(
            str(context.exception),
            "Profile already exists. Use `update_credentials` instead.",
        )

    @patch("builtins.open", new_callable=mock_open())
    @patch("runpod.cli.groups.config.functions.toml.load")
    @patch("runpod.cli.groups.config.functions.os.path.exists")
    def test_check_credentials(self, mock_exists, mock_toml_load, mock_file):
        """mock_open_call
        Tests the check_credentials function.
        """
        mock_exists.return_value = False
        passed, _ = functions.check_credentials()
        assert passed is False

        mock_exists.return_value = True
        mock_toml_load.return_value = ""
        passed, _ = functions.check_credentials()
        assert mock_file.called
        assert passed is False

        mock_exists.return_value = True
        mock_toml_load.return_value = dict({"default": "something"})
        passed, _ = functions.check_credentials()
        assert passed is False

        mock_toml_load.return_value = ValueError
        passed, _ = functions.check_credentials()
        assert passed is False

        mock_toml_load.return_value = dict({"default": "api_key"})
        passed, _ = functions.check_credentials()
        assert passed is True

    @patch("os.path.exists", return_value=True)
    @patch("runpod.cli.groups.config.functions.toml.load")
    @patch(
        "builtins.open", new_callable=mock_open, read_data='[default]\nkey = "value"'
    )
    def test_get_credentials_existing_profile(
        self, mock_open_call, mock_toml_load, mock_exists
    ):
        """
        Tests the get_credentials function.
        """
        mock_toml_load.return_value = {"default": {"key": "value"}}
        result = functions.get_credentials("default")
        assert result == {"key": "value"}
        assert mock_open_call.called
        assert mock_exists.called

    @patch("os.path.exists", return_value=True)
    @patch("runpod.cli.groups.config.functions.toml.load")
    @patch(
        "builtins.open", new_callable=mock_open, read_data='[default]\nkey = "value"'
    )
    def test_get_credentials_non_existent_profile(
        self, mock_open_call, mock_toml_load, mock_exists
    ):  # pylint: disable=line-too-long
        """
        Tests the get_credentials function.
        """
        mock_toml_load.return_value = {"default": {"key": "value"}}
        result = functions.get_credentials("non_existent")
        assert result is None
        assert mock_open_call.called
        assert mock_exists.called



""" Test the version module """

from importlib.metadata import PackageNotFoundError
from unittest.mock import patch

from runpod.version import get_version


def test_version_found():
    """Test that the version is found"""
    with patch("runpod.version.version") as mock_get_version:
        mock_get_version.return_value = "1.0.0"
        assert get_version() == "1.0.0"
        assert mock_get_version.called


def test_version_not_found():
    """Test that the version is not found"""
    with patch("runpod.version.version") as mock_get_version:
        mock_get_version.side_effect = PackageNotFoundError
        assert get_version() == "unknown"
        assert mock_get_version.called



""" RunPod | API Wrapper | Mutations | Templates """

# pylint: disable=too-many-arguments, too-many-branches


def generate_pod_template(
    name: str,
    image_name: str,
    docker_start_cmd: str = None,
    container_disk_in_gb: int = 10,
    volume_in_gb: int = None,
    volume_mount_path: str = None,
    ports: str = None,
    env: dict = None,
    is_serverless: bool = False,
    registry_auth_id: str = None,
):
    """Generate a string for a GraphQL mutation to create a new pod template."""
    input_fields = [f'name: "{name}"', f'imageName: "{image_name}"']

    # ------------------------------ Optional Fields ----------------------------- #
    if docker_start_cmd is not None:
        docker_start_cmd = docker_start_cmd.replace('"', '\\"')
        input_fields.append(f'dockerArgs: "{docker_start_cmd}"')
    else:
        input_fields.append('dockerArgs: ""')

    input_fields.append(f"containerDiskInGb: {container_disk_in_gb}")

    if volume_in_gb is not None:
        input_fields.append(f"volumeInGb: {volume_in_gb}")
    else:
        input_fields.append("volumeInGb: 0")

    if volume_mount_path is not None:
        input_fields.append(f'volumeMountPath: "{volume_mount_path}"')

    if ports is not None:
        ports = ports.replace(" ", "")
        input_fields.append(f'ports: "{ports}"')
    else:
        input_fields.append('ports: ""')

    if env is not None:
        env_string = ", ".join(
            [f'{{ key: "{key}", value: "{value}" }}' for key, value in env.items()]
        )
        input_fields.append(f"env: [{env_string}]")
    else:
        input_fields.append("env: []")

    if is_serverless:
        input_fields.append("isServerless: true")
    else:
        input_fields.append("isServerless: false")

    if registry_auth_id is not None:
        input_fields.append(f'containerRegistryAuthId : "{registry_auth_id}"')
    else:
        input_fields.append('containerRegistryAuthId : ""')

    input_fields.extend(("startSsh: true", "isPublic: false", 'readme: ""'))
    # Format the input fields into a string
    input_fields_string = ", ".join(input_fields)

    return f"""
    mutation {{
        saveTemplate(
            input: {{
                {input_fields_string}
            }}
        ) {{
            id
            name
            imageName
            dockerArgs
            containerDiskInGb
            volumeInGb
            volumeMountPath
            ports
            env {{
                key
                value
            }}
            isServerless
        }}
    }}
    """



import unittest

from runpod.serverless.utils.rp_model_cache import (
    resolve_model_cache_path_from_hugginface_repository,
)


class TestModelCache(unittest.TestCase):
    """Tests for rp_model_cache"""

    def test_with_revision(self):
        """Test with a revision"""
        path = resolve_model_cache_path_from_hugginface_repository(
            "runwayml/stable-diffusion-v1-5:experimental"
        )
        self.assertEqual(
            path, "/runpod/cache/runwayml/stable-diffusion-v1-5/experimental"
        )

    def test_without_revision(self):
        """Test without a revision"""
        path = resolve_model_cache_path_from_hugginface_repository(
            "runwayml/stable-diffusion-v1-5"
        )
        self.assertEqual(path, "/runpod/cache/runwayml/stable-diffusion-v1-5/main")

    def test_with_multiple_colons(self):
        """Test with multiple colons"""
        path = resolve_model_cache_path_from_hugginface_repository(
            "runwayml/stable-diffusion:v1-5:experimental"
        )
        self.assertEqual(
            path, "/runpod/cache/runwayml/stable-diffusion:v1-5/experimental"
        )

    def test_with_custom_path_template(self):
        """Test with a custom path template"""
        path = resolve_model_cache_path_from_hugginface_repository(
            "runwayml/stable-diffusion-v1-5:experimental",
            "/my-custom-model-cache/{model}/{revision}",
        )
        self.assertEqual(
            path, "/my-custom-model-cache/runwayml/stable-diffusion-v1-5/experimental"
        )

    def test_with_missing_model_name(self):
        """Test with a missing model name"""
        path = resolve_model_cache_path_from_hugginface_repository(":experimental")
        self.assertIsNone(path)


if __name__ == "__main__":
    unittest.main()



""" Tests for runpod.serverless.modules.rp_tips.py """

import unittest
from unittest.mock import patch

from runpod.serverless.modules.rp_tips import check_return_size


class TestTips(unittest.TestCase):
    """Tests for the Tips module"""

    @patch("runpod.serverless.modules.rp_tips.log.tip")
    def test_check_return_size_small(self, mock_log):
        """
        Tests check_return_size function with a small return_body
        """
        check_return_size("a" * 10)  # A small string

        # Ensure that the log.tip function was not called, as the return_body is small
        mock_log.assert_not_called()

    @patch("runpod.serverless.modules.rp_tips.log.tip")
    def test_check_return_size_large(self, mock_log):
        """
        Tests check_return_size function with a large return_body
        """
        check_return_size("a" * 30_000_000)  # A large string (over 20 MB)

        # Ensure that the log.tip function was called, as the return_body is large
        mock_log.assert_called()



""" Unit testing for runpod.cli.utils.rp_userspace.py """

import os
import unittest
from unittest.mock import patch

from runpod.cli.utils.rp_userspace import find_ssh_key_file


class TestFindSSHKeyFile(unittest.TestCase):
    """Unit testing for find_ssh_key_file"""

    def setUp(self):
        self.pod_ip = "127.0.0.1"
        self.pod_port = 22

    @patch("os.listdir")
    def test_no_keys_in_directory(self, mock_listdir):
        """Test find_ssh_key_file when there are no keys in the directory"""
        mock_listdir.return_value = []
        result = find_ssh_key_file(self.pod_ip, self.pod_port)
        self.assertIsNone(result)

    @patch("os.path.isfile")
    @patch("os.listdir")
    def test_all_keys_end_with_pub(self, mock_listdir, mock_isfile):
        """Test find_ssh_key_file when all keys in the directory end with .pub"""
        mock_listdir.return_value = ["key1.pub", "key2.pub"]
        mock_isfile.return_value = True
        result = find_ssh_key_file(self.pod_ip, self.pod_port)
        self.assertIsNone(result)

    @patch("paramiko.SSHClient")
    @patch("os.path.isfile")
    @patch("os.listdir")
    def test_valid_key_found(self, mock_listdir, mock_isfile, mock_ssh_client):
        """Test find_ssh_key_file when a valid key is found"""
        mock_listdir.return_value = ["key1", "key2"]
        mock_isfile.return_value = True
        mock_ssh_instance = mock_ssh_client.return_value
        mock_ssh_instance.connect.side_effect = [None, Exception("Error with key2")]
        result = find_ssh_key_file(self.pod_ip, self.pod_port)
        self.assertEqual(result, os.path.expanduser("~/.runpod/ssh/key1"))

    @patch("paramiko.SSHClient")
    @patch("os.path.isfile")
    @patch("os.listdir")
    def test_no_valid_key_found(self, mock_listdir, mock_isfile, mock_ssh_client):
        """Test find_ssh_key_file when no valid key is found"""
        mock_listdir.return_value = ["key1", "key2"]
        mock_isfile.return_value = True
        mock_ssh_instance = mock_ssh_client.return_value
        mock_ssh_instance.connect.side_effect = [
            Exception("Error with key1"),
            Exception("Error with key2"),
        ]
        result = find_ssh_key_file(self.pod_ip, self.pod_port)
        self.assertIsNone(result)



"""
runpd | error.py

This file contains the error classes for the runpod package.
"""

from typing import Optional


class RunPodError(Exception):
    """
    Base class for all runpod errors
    """

    def __init__(self, message: Optional[str] = None):
        super().__init__(message)
        self.message = message

    def __str__(self):
        if self.message:
            return self.message
        return super().__str__()


class AuthenticationError(RunPodError):
    """
    Raised when authentication fails
    """


class QueryError(RunPodError):
    """
    Raised when a GraphQL query fails
    """

    def __init__(self, message: Optional[str] = None, query: Optional[str] = None):
        super().__init__(message)
        self.query = query



"""
RunPod | CLI | Exec | Commands
"""

import click

from .functions import python_over_ssh
from .helpers import get_session_pod


@click.group("exec", help="Execute commands in a pods.")
def exec_cli():
    """A collection of CLI functions for Exec."""


@exec_cli.command("python")
@click.option("--pod_id", default=None, help="The pod ID to run the command on.")
@click.argument("file", type=click.Path(exists=True), required=True)
def remote_python(pod_id, file):
    """
    Runs a remote Python shell.
    """
    if pod_id is None:
        pod_id = get_session_pod()

    click.echo("Running remote Python shell...")
    python_over_ssh(pod_id, file)



"""
Tests for Runpod CLI exec commands.
"""

import tempfile
import unittest
from unittest.mock import patch

import click

from runpod.cli.entry import runpod_cli


class TestExecCommands(unittest.TestCase):
    """Tests for Runpod CLI exec commands."""

    def setUp(self):
        self.runner = click.testing.CliRunner()

    def test_remote_python_with_provided_pod_id(self):
        """Tests the remote_python command when pod_id is provided directly."""
        with tempfile.NamedTemporaryFile() as temp_file, patch(
            "runpod.cli.groups.exec.commands.python_over_ssh"
        ) as mock_python_over_ssh:
            result = self.runner.invoke(
                runpod_cli,
                ["exec", "python", "--pod_id", "sample_pod_id", temp_file.name],
            )
            assert result.exit_code == 0
            mock_python_over_ssh.assert_called_with("sample_pod_id", temp_file.name)

    def test_remote_python_without_provided_pod_id_stored(self):
        """Tests the remote_python command when pod_id is retrieved from storage."""
        with tempfile.NamedTemporaryFile() as temp_file, patch(
            "runpod.cli.groups.exec.commands.python_over_ssh"
        ) as mock_python_over_ssh, patch(
            "runpod.cli.groups.exec.commands.get_session_pod",
            return_value="stored_pod_id",
        ) as mock_get_pod_id:  # pylint: disable=line-too-long
            mock_python_over_ssh.return_value = None
            result = self.runner.invoke(runpod_cli, ["exec", "python", temp_file.name])
            assert result.exit_code == 0
            mock_get_pod_id.assert_called_once()
            mock_python_over_ssh.assert_called_with("stored_pod_id", temp_file.name)

    def test_remote_python_without_provided_pod_id_prompt(self):
        """Tests the remote_python command when pod_id is prompted to user."""
        with tempfile.NamedTemporaryFile() as temp_file, patch(
            "runpod.cli.groups.exec.commands.python_over_ssh"
        ) as mock_python_over_ssh, patch(
            "runpod.cli.groups.exec.commands.get_session_pod",
            side_effect=lambda: click.prompt(
                "Please provide the pod ID", "prompted_pod_id"
            ),
        ) as mock_get_pod_id:  # pylint: disable=line-too-long
            mock_python_over_ssh.return_value = None
            result = self.runner.invoke(runpod_cli, ["exec", "python", temp_file.name])
            assert result.exit_code == 0
            mock_get_pod_id.assert_called_once()
            mock_python_over_ssh.assert_called_with("prompted_pod_id", temp_file.name)



"""
Watches a directory for changes and syncs them to a remote directory.
"""

import threading
import time

from watchdog.events import FileSystemEventHandler
from watchdog.observers.polling import PollingObserver as Observer

from runpod.cli import STOP_EVENT

from .rp_runpodignore import get_ignore_list, should_ignore


class WatcherHandler(FileSystemEventHandler):
    """Watches a directory for changes and syncs them to a remote directory."""

    def __init__(self, action_function, local_path):
        self.action_function = action_function
        self.local_path = local_path

        self.ignore_list = get_ignore_list()
        self.debouncer = None

    def on_any_event(self, event):
        """Called on any event."""
        if event.is_directory or should_ignore(event.src_path, self.ignore_list):
            return

        if self.debouncer is not None:
            self.debouncer.cancel()  # Cancel any existing timer

        # Start a new timer that will call the action function after 1 second
        self.debouncer = threading.Timer(0.5, self.action_function)
        self.debouncer.start()


def start_watcher(action_function, local_path):
    """
    Starts the watcher.
    """
    event_handler = WatcherHandler(action_function, local_path)
    observer = Observer()
    observer.schedule(event_handler, local_path, recursive=True)
    observer.start()

    try:
        while not STOP_EVENT.is_set():
            time.sleep(0.5)
    finally:
        observer.stop()
        observer.join()


def sync_directory(ssh_client, local_path, remote_path):
    """
    Syncs a local directory to a remote directory.
    """

    def sync():
        print("Syncing files...")
        ssh_client.rsync(local_path, remote_path, quiet=True)

    threading.Thread(target=start_watcher, daemon=True, args=(sync, local_path)).start()

    return sync  # For testing purposes



""" Allows the CLI to be imported as a module. """

import threading

from .groups import config, ssh

STOP_EVENT = threading.Event()


# --------------------------- runpod.toml Defaults --------------------------- #
BASE_DOCKER_IMAGE = "runpod/base:0.4.3-cuda{cuda_version}"
GPU_TYPES = [
    "NVIDIA RTX A4000",
    "NVIDIA RTX A4500",
    "NVIDIA RTX A5000",
    "NVIDIA GeForce RTX 3090",
    "NVIDIA RTX A6000",
]
ENV_VARS = {
    "POD_INACTIVITY_TIMEOUT": "120",
    "RUNPOD_DEBUG_LEVEL": "debug",
    "UVICORN_LOG_LEVEL": "warning",
}



""" Tests for runpod.serverless.modules.rp_fastapi.py """

# pylint: disable=protected-access

import asyncio
import os
import unittest
from unittest.mock import MagicMock, Mock, patch

import pytest
import requests

import runpod
from runpod.serverless.modules import rp_fastapi


class TestFastAPI(unittest.TestCase):
    """Tests the FastAPI"""

    def setUp(self) -> None:
        self.handler = Mock()
        self.handler.return_value = {"result": "success"}

        self.error_handler = Mock()
        self.error_handler.side_effect = Exception("test error")

    def test_start_serverless_with_realtime(self):
        """
        Tests the start_serverless() method with the realtime option.
        """
        module_location = "runpod.serverless.modules.rp_fastapi"
        with patch(
            f"{module_location}.Heartbeat.start_ping", Mock()
        ) as mock_ping, patch(
            f"{module_location}.FastAPI", Mock()
        ) as mock_fastapi, patch(
            f"{module_location}.APIRouter", return_value=Mock()
        ) as mock_router, patch(
            f"{module_location}.uvicorn", Mock()
        ) as mock_uvicorn:

            rp_fastapi.RUNPOD_REALTIME_PORT = "1111"
            rp_fastapi.RUNPOD_ENDPOINT_ID = "test_endpoint_id"

            os.environ["RUNPOD_REALTIME_PORT"] = "1111"
            os.environ["RUNPOD_ENDPOINT_ID"] = "test_endpoint_id"

            runpod.serverless.start({"handler": self.handler})

            self.assertTrue(mock_ping.called)

            self.assertTrue(mock_fastapi.called)
            self.assertTrue(mock_router.called)

            self.assertTrue(rp_fastapi.RUNPOD_ENDPOINT_ID == "test_endpoint_id")
            self.assertTrue(mock_router.return_value.add_api_route.called)

            self.assertTrue(mock_uvicorn.run.called)

            os.environ.pop("RUNPOD_REALTIME_PORT")
            os.environ.pop("RUNPOD_ENDPOINT_ID")

    def test_webhook_sender_success(self):
        """Test the webhook sender when the request is successful."""
        module_location = "runpod.serverless.modules.rp_fastapi.requests.Session.post"

        with patch(module_location, new_callable=MagicMock) as mock_post:
            # Simulate a successful response
            mock_post.return_value.status_code = 200

            # Call the function
            success = rp_fastapi._send_webhook("test_webhook", {"test": "output"})
            assert success is True

    def test_webhook_sender_failure(self):
        """Test the webhook sender when the request fails."""
        module_location = "runpod.serverless.modules.rp_fastapi.requests.Session.post"

        with patch(module_location, new_callable=MagicMock) as mock_post:
            # Configure the mock to simulate a failure (e.g., a 500 status code)
            mock_post.return_value.raise_for_status.side_effect = requests.HTTPError()
            mock_post.return_value.status_code = 500

            # Call the function
            success = rp_fastapi._send_webhook("test_webhook", {"test": "output"})
            assert success is False

    @pytest.mark.asyncio
    def test_run(self):
        """
        Tests the _run() method.
        """
        loop = asyncio.get_event_loop()

        module_location = "runpod.serverless.modules.rp_fastapi"
        with patch(
            f"{module_location}.Heartbeat.start_ping", Mock()
        ) as mock_ping, patch(f"{module_location}.FastAPI", Mock()), patch(
            f"{module_location}.APIRouter", return_value=Mock()
        ), patch(
            f"{module_location}.uvicorn", Mock()
        ), patch(
            f"{module_location}.uuid.uuid4", return_value="123"
        ):

            job_object = rp_fastapi.Job(
                id="test_job_id", input={"test_input": "test_input"}
            )

            default_input_object = rp_fastapi.DefaultRequest(
                input={"test_input": "test_input"}
            )

            # Test with handler
            worker_api = rp_fastapi.WorkerAPI({"handler": self.handler})

            run_return = asyncio.run(worker_api._realtime(job_object))
            assert run_return == {"output": {"result": "success"}}

            debug_run_return = asyncio.run(worker_api._sim_run(default_input_object))
            assert debug_run_return == {"id": "test-123", "status": "IN_PROGRESS"}

            self.assertTrue(mock_ping.called)

            # Test with generator handler
            def generator_handler(job):
                del job
                yield {"result": "success"}

            generator_worker_api = rp_fastapi.WorkerAPI({"handler": generator_handler})
            generator_run_return = asyncio.run(
                generator_worker_api._sim_run(default_input_object)
            )
            assert generator_run_return == {"id": "test-123", "status": "IN_PROGRESS"}

        loop.close()

    @pytest.mark.asyncio
    def test_runsync(self):
        """
        Tests the _runsync() method.
        """
        loop = asyncio.get_event_loop()

        module_location = "runpod.serverless.modules.rp_fastapi"
        with patch(f"{module_location}.FastAPI", Mock()), patch(
            f"{module_location}.APIRouter", return_value=Mock()
        ), patch(f"{module_location}.uvicorn", Mock()), patch(
            f"{module_location}.uuid.uuid4", return_value="123"
        ), patch(
            f"{module_location}.threading"
        ) as mock_threading:

            default_input_object = rp_fastapi.DefaultRequest(
                input={"test_input": "test_input"}
            )

            input_object_with_webhook = rp_fastapi.DefaultRequest(
                input={"test_input": "test_input"}, webhook="test_webhook"
            )

            # Test with handler
            worker_api = rp_fastapi.WorkerAPI({"handler": self.handler})

            runsync_return = asyncio.run(worker_api._sim_runsync(default_input_object))
            assert runsync_return == {
                "id": "test-123",
                "status": "COMPLETED",
                "output": {"result": "success"},
            }

            # Test with generator handler
            def generator_handler(job):
                del job
                yield {"result": "success"}

            generator_worker_api = rp_fastapi.WorkerAPI({"handler": generator_handler})
            generator_runsync_return = asyncio.run(
                generator_worker_api._sim_runsync(default_input_object)
            )
            assert generator_runsync_return == {
                "id": "test-123",
                "status": "COMPLETED",
                "output": [{"result": "success"}],
            }

            # Test with error handler
            error_worker_api = rp_fastapi.WorkerAPI({"handler": self.error_handler})
            error_runsync_return = asyncio.run(
                error_worker_api._sim_runsync(default_input_object)
            )
            assert "error" in error_runsync_return

            # Test webhook caller sent
            asyncio.run(worker_api._sim_runsync(input_object_with_webhook))
            assert mock_threading.Thread.called

        loop.close()

    @pytest.mark.asyncio
    def test_stream(self):
        """
        Tests the _stream() method.
        """
        loop = asyncio.get_event_loop()

        module_location = "runpod.serverless.modules.rp_fastapi"
        with patch(f"{module_location}.FastAPI", Mock()), patch(
            f"{module_location}.APIRouter", return_value=Mock()
        ), patch(f"{module_location}.uvicorn", Mock()), patch(
            f"{module_location}.uuid.uuid4", return_value="123"
        ), patch(
            f"{module_location}.threading"
        ) as mock_threading:

            default_input_object = rp_fastapi.DefaultRequest(
                input={"test_input": "test_input"}
            )

            input_object_with_webhook = rp_fastapi.DefaultRequest(
                input={"test_input": "test_input"}, webhook="test_webhook"
            )

            worker_api = rp_fastapi.WorkerAPI({"handler": self.handler})

            # Add job to job_list
            asyncio.run(worker_api._sim_run(default_input_object))

            stream_return = asyncio.run(worker_api._sim_stream("test_job_id"))
            assert stream_return == {
                "id": "test_job_id",
                "status": "FAILED",
                "error": "Job ID not found",
            }

            stream_return = asyncio.run(worker_api._sim_stream("test-123"))
            assert stream_return == {
                "id": "test-123",
                "status": "FAILED",
                "error": "Stream not supported, handler must be a generator.",
            }

            # Test with generator handler
            def generator_handler(job):
                del job
                yield {"result": "success"}

            generator_worker_api = rp_fastapi.WorkerAPI({"handler": generator_handler})
            generator_stream_return = asyncio.run(
                generator_worker_api._sim_stream("test-123")
            )
            assert generator_stream_return == {
                "id": "test-123",
                "status": "COMPLETED",
                "stream": [{"output": {"result": "success"}}],
            }

            # Test webhook caller sent
            asyncio.run(generator_worker_api._sim_run(input_object_with_webhook))
            asyncio.run(generator_worker_api._sim_stream("test-123"))
            assert mock_threading.Thread.called

        loop.close()

    @pytest.mark.asyncio
    def test_status(self):
        """
        Tests the _status() method.
        """
        loop = asyncio.get_event_loop()

        module_location = "runpod.serverless.modules.rp_fastapi"
        with patch(f"{module_location}.FastAPI", Mock()), patch(
            f"{module_location}.APIRouter", return_value=Mock()
        ), patch(f"{module_location}.uvicorn", Mock()), patch(
            f"{module_location}.uuid.uuid4", return_value="123"
        ), patch(
            f"{module_location}.threading"
        ) as mock_threading:

            worker_api = rp_fastapi.WorkerAPI({"handler": self.handler})

            default_input_object = rp_fastapi.DefaultRequest(
                input={"test_input": "test_input"}
            )

            input_object_with_webhook = rp_fastapi.DefaultRequest(
                input={"test_input": "test_input"}, webhook="test_webhook"
            )

            # Add job to job_list
            asyncio.run(worker_api._sim_run(default_input_object))

            status_return = asyncio.run(worker_api._sim_status("test_job_id"))
            assert status_return == {
                "id": "test_job_id",
                "status": "FAILED",
                "error": "Job ID not found",
            }

            status_return = asyncio.run(worker_api._sim_status("test-123"))
            assert status_return == {
                "id": "test-123",
                "status": "COMPLETED",
                "output": {"result": "success"},
            }

            # Test webhook caller sent
            asyncio.run(worker_api._sim_run(input_object_with_webhook))
            asyncio.run(worker_api._sim_status("test-123"))
            assert mock_threading.Thread.called

            # Test with generator handler
            def generator_handler(job):
                del job
                yield {"result": "success"}

            generator_worker_api = rp_fastapi.WorkerAPI({"handler": generator_handler})
            asyncio.run(generator_worker_api._sim_run(default_input_object))
            generator_stream_return = asyncio.run(
                generator_worker_api._sim_status("test-123")
            )
            assert generator_stream_return == {
                "id": "test-123",
                "status": "COMPLETED",
                "output": [{"result": "success"}],
            }

            # Test with error handler
            error_worker_api = rp_fastapi.WorkerAPI({"handler": self.error_handler})
            asyncio.run(error_worker_api._sim_run(default_input_object))
            error_status_return = asyncio.run(error_worker_api._sim_status("test-123"))
            assert "error" in error_status_return

        loop.close()



""" Example of creating a template with the Runpod API. """

import runpod

# Set your global API key with `runpod config` or uncomment the line below:
# runpod.api_key = "YOUR_RUNPOD_API_KEY"

try:

    new_template = runpod.create_template(name="test", image_name="runpod/base:0.1.0")

    print(new_template)

except runpod.error.QueryError as err:
    print(err)
    print(err.query)



"""
Handles getting stuff from environment variables and updating the global state like job id.
"""

import os
import time
import uuid
from typing import Any, Dict, Optional

from .rp_logger import RunPodLogger


log = RunPodLogger()

REF_COUNT_ZERO = time.perf_counter()  # Used for benchmarking with the debugger.

WORKER_ID = os.environ.get("RUNPOD_POD_ID", str(uuid.uuid4()))


# ----------------------------------- Flags ---------------------------------- #
IS_LOCAL_TEST = os.environ.get("RUNPOD_WEBHOOK_GET_JOB", None) is None


# ------------------------------- Job Tracking ------------------------------- #
class Job:
    """
    Represents a job object.

    Args:
        job_id: The id of the job, a unique string.
        job_input: The input to the job.
        webhook: The webhook to send the job output to.
    """

    def __init__(
        self,
        id: str,
        input: Optional[Dict[str, Any]] = None,
        webhook: Optional[str] = None,
        **kwargs
    ) -> None:
        self.id = id
        self.input = input
        self.webhook = webhook

        for key, value in kwargs.items():
            setattr(self, key, value)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Job):
            return self.id == other.id
        return False

    def __hash__(self) -> int:
        return hash(self.id)

    def __str__(self) -> str:
        return self.id


# ---------------------------------------------------------------------------- #
#                                    Tracker                                   #
# ---------------------------------------------------------------------------- #
class JobsProgress(set):
    """Track the state of current jobs in progress."""

    _instance = None

    def __new__(cls):
        if JobsProgress._instance is None:
            JobsProgress._instance = set.__new__(cls)
        return JobsProgress._instance

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}>: {self.get_job_list()}"

    def clear(self) -> None:
        return super().clear()

    def add(self, element: Any):
        """
        Adds a Job object to the set.

        If the added element is a string, then `Job(id=element)` is added
        
        If the added element is a dict, that `Job(**element)` is added
        """
        if isinstance(element, str):
            element = Job(id=element)

        if isinstance(element, dict):
            element = Job(**element)

        if not isinstance(element, Job):
            raise TypeError("Only Job objects can be added to JobsProgress.")

        return super().add(element)

    def remove(self, element: Any):
        """
        Removes a Job object from the set.

        If the element is a string, then `Job(id=element)` is removed
        
        If the element is a dict, then `Job(**element)` is removed
        """
        if isinstance(element, str):
            element = Job(id=element)

        if isinstance(element, dict):
            element = Job(**element)

        if not isinstance(element, Job):
            raise TypeError("Only Job objects can be removed from JobsProgress.")

        return super().discard(element)

    def get(self, element: Any) -> Job:
        if isinstance(element, str):
            element = Job(id=element)

        if not isinstance(element, Job):
            raise TypeError("Only Job objects can be retrieved from JobsProgress.")

        for job in self:
            if job == element:
                return job

    def get_job_list(self) -> str:
        """
        Returns the list of job IDs as comma-separated string.
        """
        if not len(self):
            return None

        return ",".join(str(job) for job in self)

    def get_job_count(self) -> int:
        """
        Returns the number of jobs.
        """
        return len(self)
